(library (ast-bubbles)
  (export make-AST-bubbles AST-bubbles-intersect)
  (import (except (rnrs) case cond if) (safe-forms) (myers-stacks)
          (tree-locations) (ast) (idents))

;; KEY TRICK: the bubble algorithm creates the bubble lists in order
;; sorted by their ident-number
(define (AST-bubbles-intersect set1 set2)
  (define < ident>?)
  (define rec AST-bubbles-intersect)
  (cond
   [(or (null? set1) (null? set2)) '()]
   [(< (car set1) (car set2)) (rec (cdr set1) set2)]
   [(< (car set2) (car set1)) (rec set1 (cdr set2))]
   [else (cons (car set1) (rec (cdr set1) (cdr set2)))]))

(define (make-AST-bubbles ast)
  (define tree-location-table
    (make-tree-location-table
     ast
     AST-tree-start-location-set! AST-tree-end-location-set!
     ident-asts ident-asts-set!
     ident-locations ident-locations-set!
     contains mark add
     (lambda (ast)
       (AST-case ast
         [(Ref ident) (mark ast ident)]
         [(Quote obj) (values)]
         [(Primref safe name) (values)]
         ;; TODO: how to mark non-leaf (for now not needed b/c mutable vars)
         [(Set ident expr) (contains expr)]
         [(App fun args) (contains fun) (vector-for-each contains args)]
         [(If test conseq altern)
          (contains test) (contains conseq) (contains altern)]
         [(CaseLambda table clauses)
          (letrec ([do-clause
                    (lambda (clause)
                      (for-each add (case-lambda-clause-formals&rest clause))
                      (contains (case-lambda-clause-body clause)))])
            (vector-for-each do-clause clauses))]
         [(Seq e1 e2) (contains e1) (contains e2)]
         [(Letrec bindings body)
          (vector-for-each add      (vector-map letrec-binding-lhs bindings))
          (vector-for-each contains (vector-map letrec-binding-rhs bindings))
          (contains body)])
       (values))))

  (define (AST-bubble-vars-add! ast ident)
    (AST-bubble-vars-set! ast (cons ident (AST-bubble-vars ast))))

  (define (AST-bubble-vars-contains? ast ident)
    (let ([bubbles (AST-bubble-vars ast)])
      (and (not (null? bubbles)) (ident=? ident (car bubbles)))))

  (define (bubble-base tpt ast ident)
    (define context (AST-context ast))
    (define locations (ident-locations ident))
    (define asts (ident-asts ident))
    (define prev-ref
      (tree-location-table-prev-ref
       tpt ast ident #f (AST-tree-start-location ast) locations asts))
    (define prev-base
      (myers-stack-eq-suffix
       context (if prev-ref (AST-context prev-ref) myers-stack-null)))
    (define next-ref
      (tree-location-table-next-ref
       tpt ast ident #f (AST-tree-end-location ast) locations asts))
    (define next-base
      (myers-stack-eq-suffix
       context (if next-ref (AST-context next-ref) myers-stack-null)))

    (if (> (myers-stack-length prev-base) (myers-stack-length next-base))
        prev-base
        next-base))

  ;; inflate - move up context and stop just before would
  ;;           include extra ref to ident
  (define (inflate queue ast ident)
    (let ([base (bubble-base tree-location-table ast ident)]) ;; length >= 0
      ;; TODO: if not found (then add nothing)
      ;; (instead of or in additon to "if stack-null-base")
      (if (myers-stack-null? base)
          queue
          (let* ([context (AST-context ast)] ;; length >= 1
                 [root (myers-stack-car (myers-stack-previous context base))])
            (if (AST-bubble-vars-contains? root ident)
                queue
                (begin
                  ;; mark root
                  (AST-bubble-vars-add! root ident)
                  ;; add parent of root to queue
                  (cons (myers-stack-car base) queue)))))))

  ;; Give the variables a number so we can keep them sorted
  (let loop ([counter 0]
             [idents (tree-location-table-key-list tree-location-table)])
    (unless (null? idents)
      (ident-number-set! (car idents) counter)
      (loop (+ 1 counter) (cdr idents))))

  ;; NOTE: each ident is separate passes so AST-bubble-vars-add!
  ;; keeps the list uniqified
  (for-each
   (lambda (ident)
     (when (not (ident-assigned ident))
           ;; put all occurances in the work queue
           (let loop ([queue (vector->list (ident-asts ident))])
             (unless (null? queue) ;; loop until queue empty
               (loop (inflate (cdr queue) (car queue) ident))))))
   ;; order is critical to get the sets ordered
   (tree-location-table-key-list tree-location-table)))

)
