(library (ast-unparse)
 (export AST->Trex)
 (import (rnrs) (srfi-39) (flow-graphs) (ast) (types) (value-flow-graphs)
         (prefix (tprogram) Trex:))

;; prim type table -- duplicates knowledge in prims
(define mark-safe?
  (let () ;; deeded to allow define
    (define (arity-match? arity actual-types)
      (if (>= arity 0)
          (= arity (length actual-types))
          (>= (+ (- arity) 1) (length actual-types))))
    (define (type-match? type actual-type) 
      (let ([type (maybe-non-proc-type->type type)])
        (or (type=? type top-type) (type<=? actual-type type))))
    (define (type-with-rest-match? types actual-types)
      (let f ([types types] [actual-types actual-types])
        (if (or (null? actual-types) (null? types))
            #t
            (let ([type (car types)] [remaining-types (cdr types)])
              (and
               (type<=? (car actual-types) (maybe-non-proc-type->type type))
               (if (null? remaining-types)
                   (f types (cdr actual-types))
                   (f remaining-types (cdr actual-types))))))))
    (define (types-match? sig actual-types)
      (let ([arity (car sig)] [types (cdr sig)])
        (if (>= arity 0)
            (for-all type-match? types actual-types)
            (type-with-rest-match? types actual-types))))
    (define-syntax build-prim-entry
      (syntax-rules ()
        [(_ ht (prim* ...) ?value)
         (let ([value ?value])
           (hashtable-set! ht 'prim* value) ...)]
        [(_ ht prim value) (hashtable-set! ht 'prim value)]))
    (define-syntax define-prim-table 
      (lambda (x)
        (define (ellipsis? x)
            (and (identifier? x) (free-identifier=? x #'(... ...))))
        (define (find-arity arg*)
            (syntax-case arg* ()
              [(arg arg* ... dots)
               (ellipsis? #'dots)
               #`(#,(fx- -1 (length #'(arg arg* ...))) arg arg* ...)]
              [(arg* ...) #`(#,(length #'(arg* ...)) arg* ...)]))
        (syntax-case x (sig)
          [(_ name (name* [sig arg** ...] extras** ...) ...)
           (with-syntax ([(((arity** arg*** ...) ...) ...)
                          (map (lambda (arg**)
                                 (map find-arity arg**))
                               #'((arg** ...) ...))])
             #'(define name
                 (let ([ht (make-eq-hashtable)])
                   (build-prim-entry
                    ht name* (list (list (list arity** arg*** ...) ...)
                                   (list extras** ...))) ...
                   ht)))])))
    (define-prim-table prim-table
      ((car cdr) [sig (pair-type-bits)])
      ((set-car! set-cdr!) [sig (pair-type-bits top-type)])
      (make-vector [sig (length-type-bits) (length-type-bits top-type)])
      (vector-length [sig (vector-type-bits)])
      (vector-ref [sig (vector-type-bits length-type-bits)]
        (let ([vector-type (maybe-non-proc-type->type vector-type-bits)]
              [length-type (maybe-non-proc-type->type length-type-bits)])
          (lambda (src name types)
            (if (and (= (length types) 2) (type<=? (car types) vector-type))
                (if (type<=? (cadr types) length-type)
                    (Trex:Primref src #t '$vector-check-range-ref)
                    (Trex:Primref src #t '$vector-check-fx-range-ref))
                (Trex:Primref src #t name)))))
      (vector-set! [sig (vector-type-bits length-type-bits top-type)]
        (let ([vector-type (maybe-non-proc-type->type vector-type-bits)]
              [length-type (maybe-non-proc-type->type length-type-bits)])
          (lambda (src name types)
            (if (and (= (length types) 3) (type<=? (car types) vector-type))
                (if (type<=? (cadr types) length-type)
                    (Trex:Primref src #t '$vector-check-range-set!)
                    (Trex:Primref src #t '$vector-check-fx-range-set!))
                (Trex:Primref src #t name)))))
      ((fldiv flmod fldiv0 flmod0 flexpt) [sig (flonum-type-bits flonum-type-bits)])
      ((flatan fllog) [sig (flonum-type-bits) (flonum-type-bits flonum-type-bits)])
      ((flsqrt flabs flnumerator fldenominator flfloor flceiling fltruncate
        flround flexp flsin flcos fltan flasin flacos)
       [sig (flonum-type-bits)])
      ((fl+ fl*) [sig (flonum-type-bits ...)])
      ((fl- fl/ flmax flmin)
       [sig (flonum-type-bits) (flonum-type-bits flonum-type-bits ...)])
      ((flpositive? flnegative? flzero? flinteger?
        flodd? fleven?  flfinite? flinfinite? flnan?)
       [sig (flonum-type-bits)])
      ((fl>? fl<? fl>=? fl<=? fl=?)
       [sig (flonum-type-bits flonum-type-bits flonum-type-bits ...)])
      (fxxor [sig (fxnum-type-bits ...)])
      ((fx+ fx*) [sig (fxnum-type-bits ...)])
      ((fx- fx/) [sig (fxnum-type-bits fxnum-type-bits ...)])
      ((fxarithmetic-shift-left fxarithmetic-shift-right fxarithmetic-shift)
       [sig (fxnum-type-bits fxnum-type-bits)])
      ((fx=? fx<=? fx>=? fx<? fx>?)
       [sig (fxnum-type-bits fxnum-type-bits fxnum-type-bits ...)])
      ((fx<= fx< fx= fx>= fx>) [sig (fxnum-type-bits fxnum-type-bits ...)])
      ((char=? char<=? char<? char>? char>=?)
       [sig (char-type-bits char-type-bits ...)])
      ((r6rs:char=? r6rs:char<=? r6rs:char<? r6rs:char>? r6rs:char>=?)
       [sig (char-type-bits char-type-bits char-type-bits ...)])
      (integer->char [sig (fxnum-type-bits)])
      (char->integer [sig (char-type-bits)])
      (string-length [sig (string-type-bits)])
      (string [sig (char-type-bits ...)])
      (make-string [sig (fxnum-type-bits) (fxnum-type-bits char-type-bits)])
      ;; can do better with this, need to handle like vector-set! and vector-ref
      (string-set! [sig (string-type-bits fxnum-type-bits char-type-bits)])
      (string-ref [sig (string-type-bits fxnum-type-bits)]))
    (lambda (src prim actual-types)
      (define (check-sig sig)
        (and (arity-match? (car sig) actual-types)
             (types-match? sig actual-types)))
      (define (check-prim name sigs extras)
          (cond
           [(and (pair? extras) (procedure? (car extras)))
            ((car extras) src name actual-types)]
           [(not (null? extras)) #f]
           [(exists check-sig sigs) (Trex:Primref src #f name)]
           [else #f]))
      (cond
       [(hashtable-ref prim-table prim #f) =>
        (lambda (v) (apply check-prim prim v))]
       [else #f]))))

(define (map-values vals f ls . more)
    (define (return-values vals vals*)
      (if (null? vals*)
          (apply values (map list vals))
          (apply values (map cons vals vals*))))
    (if (null? more)
        (let map1 ([ls ls])
          (if (null? ls)
              (apply values vals)
              (let-values ([vals (f (car ls))] [vals* (map1 (cdr ls))])
                (return-values vals vals*))))
        (let map-more ([ls ls] [more more])
          (if (null? ls)
              (apply values vals)
              (let-values ([vals (apply f (car ls) (map car more))]
                           [vals* (map-more (cdr ls) (map cdr more))])
                (return-values vals vals*))))))

(define (AST-output-type ast) (flow-graph-node-output (AST-output-node ast)))

(define (AST->Trex ast)
  (define changed (make-parameter #f))

  (define (rec ast)
    (define src (AST-src ast))
    (values
     (AST-case ast
       ;; Things that don't collapse
       [(Ref id) (Trex:Ref src id)]
       [(Quote obj) (Trex:Quote src obj)]
       [(Primref safe name) (Trex:Primref src safe name)]
       [(Set id body) (let-values ([(body body-type) (rec body)])
                        (Trex:Set src id body))]
       [(App fun actuals) 
        (let-values ([(actuals actual-types)
                      (map-values (list '() '()) rec (vector->list actuals))])
          (AST-case fun
            [(Primref safe name)
             (cond
              [(and safe (mark-safe? (AST-src fun) name actual-types)) =>
               (lambda (fun) (Trex:App src fun actuals))]
              [else
               (let-values ([(fun fun-type) (rec fun)])
                 (Trex:App src fun actuals))])]
            [else
             (let-values ([(fun fun-type) (rec fun)])
               (Trex:App src fun actuals))]))]
       [(Letrec bindings body)
        (let ([bindings (vector->list bindings)])
          (let-values ([(body type) (rec body)]
                       [(rhs rhs-types)
                        (map-values (list '() '()) rec
                                    (map letrec-binding-rhs bindings))])
            (Trex:Letrec src (map letrec-binding-lhs bindings) rhs body)))]
       [(CaseLambda table clauses)
        (let ([Clause->Trex
               (lambda (clause)
                 (let-values ([(body type) (rec (case-lambda-clause-body clause))])
                   (Trex:CaseLambdaClause
                    (append (vector->list (case-lambda-clause-formals clause))
                            (or (case-lambda-clause-rest clause) '()))
                    body)))]
              [Clause-called
               (lambda (clause)
                 (flow-graph-node-output
                  (clause-info-in-called (case-lambda-clause-info clause))))])
          (Trex:CaseLambda
           src (map Clause->Trex
                    (filter Clause-called (vector->list clauses)))))]
    
       ;; Things that might collapse
       [(If test conseq altern)
        (let-values ([(test type) (rec test)]
                     [(conseq c-type) (rec conseq)]
                     [(altern a-type) (rec altern)])
          (let ([conseq-reachable (true-type? type)]
                [altern-reachable (false-type? type)])
            (cond
             [(and conseq-reachable altern-reachable)
              (Trex:If src test conseq altern)]
             [conseq-reachable (changed #t) (Trex:Seq src test conseq)]
             [altern-reachable (changed #t) (Trex:Seq src test altern)]
             [else (changed #t) test])))]
       [(Seq expr1 expr2)
        (let-values ([(expr1 e1-type) (rec expr1)]
                     [(expr2 e2-type) (rec expr2)])
          (if (non-bot-type? e1-type)
              (Trex:Seq src expr1 expr2)
              (begin (changed #t) expr1)))])
     (AST-output-type ast)))

  (let ([trex
         (AST-case ast
           [(App fun actuals)
            (AST-case fun
              [(Primref safe name) ;; assumes "escape" is always unsafe
               (cond
                [(eq? name 'escape)
                 (cond
                  [(= 1 (vector-length actuals))
                   (let-values ([(body type) (rec (vector-ref actuals 0))])
                     body)]
                  [else #f])]
                [else #f])]
              [else #f])]
           [else #f])])
    (or trex (assertion-violation
              'AST->Trex "Incorrect toplevel program expression" ast))))

)
