(library (ast)
  (export
   AST? AST-case Ref Quote Primref Set App If CaseLambda Seq Letrec

   Trex->AST

   AST-src ;; obj
   AST-context ;; myers-stack of AST
   AST-output-node ;; flow-graph-node of type
   AST-output-node-set!
   AST-graph-context ;; composition-stack of env-flow-functions
   AST-graph-context-set!
   AST-graph-context-thunk ;; (lambda () ...)
   AST-graph-context-thunk-set!
   AST-contained-nodes ;; list of flow-graph-node
   AST-contained-nodes-set!
   AST-bubble-vars ;; list of ident
   AST-bubble-vars-set!
   AST-tree-start-location ;; int
   AST-tree-start-location-set!
   AST-tree-end-location ;; int
   AST-tree-end-location-set!

   case-lambda-clause?
   case-lambda-clause-ref
   case-lambda-clause-formals case-lambda-clause-rest
   case-lambda-clause-formals&rest case-lambda-clause-body
   case-lambda-clause-info case-lambda-clause-info-set!
   case-lambda-clause-contained-nodes case-lambda-clause-contained-nodes-set!

   letrec-binding-lhs letrec-binding-rhs
   )

  (import (rnrs) (prefix (tprogram) Trex:)
          (idents) (srfi-39) (datatype) (myers-stacks)
          (only (composition-stacks) composition-stack?)
          (only (flow-graphs) flow-graph-node?))

(define-datatype
  (AST (src #f)
       (output-node #f flow-graph-node?)
       (contained-nodes '() #;(list-of flow-graph-node?))
       (context #f myers-stack?) ;; myers stack of ast-nodes including self
       (graph-context #f composition-stack?)
       (graph-context-thunk #f #;(pair flow-graph-node? (list-of ast)))
       (bubble-vars '() #;(list-of ident?))
       (tree-start-location #f integer?)
       (tree-end-location #f integer?)
       )

  (Ref (id ident?))
  (Quote obj)
  (Primref (safe boolean?) (name symbol?))
  (Set (id ident?) (expr AST?))
  (App (expr AST?) (expr* (vector-of AST?)))
  (If (test AST?) (thn AST?) (els AST?))
  (CaseLambda (table (vector-of (lambda (x) (or (not x) (integer? x)))))
              (clause* (vector-of case-lambda-clause?)))
  (Seq (expr1 AST?) (expr2 AST?))
  (Letrec (binding* (vector-of letrec-binding?)) (body AST?)))

(define-record-type letrec-binding
  (nongenerative)
  (fields lhs rhs)
  (protocol
   (lambda (make-letrec-binding)
     (lambda (lhs rhs)
       (assert (ident? lhs))
       (assert (AST? rhs))
       (make-letrec-binding lhs rhs)))))

(define-record-type case-lambda-clause
  (nongenerative)
  (fields formals rest body formals&rest
          (mutable info) (mutable contained-nodes))
  (protocol
   (lambda (make)
     (lambda (formals rest body)
       (assert ((vector-of ident?) formals))
       (assert (or (not rest) (ident? rest)))
       (assert (AST? body))
       (make formals rest body
             (append (vector->list formals) (if rest (list rest) '())) #f '())))))

;; vector[case-lambda-clause] -> vector[#f or clause-index]
;;   last slot is for arguments beyond vector-length
(define (make-case-lambda-table clauses)
  (define (vector-set-or! vector index value)
    (vector-set! vector index (or (vector-ref vector index) value)))
  (define has-rest (vector-map case-lambda-clause-rest clauses))
  (define num-non-rest-args
    (vector-map vector-length
                (vector-map case-lambda-clause-formals clauses)))
  (define table-length (+ 2 (apply max (cons 0 (vector->list num-non-rest-args)))))
  (define table (make-vector table-length #f))
  (let clause-loop ([clause-number 0][min-rest +inf.0])
    (unless (= clause-number (vector-length clauses))
      (let ([table-index (vector-ref num-non-rest-args clause-number)])
        (if (not (vector-ref has-rest clause-number))
            (begin
              (vector-set-or! table table-index clause-number)
              (clause-loop (+ 1 clause-number) min-rest))
            (begin
              (let rest-loop ([table-index table-index])
                (unless (or (>= table-index min-rest)
                            (>= table-index table-length))
                  (vector-set-or! table table-index clause-number)
                  (rest-loop (+ 1 table-index))))
              (clause-loop (+ 1 clause-number) (min min-rest table-index)))))))
  table)

;; case-lambda-table * vector[case-lambda-clause] * int -> #f or case-lambda-clause
(define (case-lambda-clause-ref table clauses arg-count)
   (let ([index (vector-ref table (min arg-count (- (vector-length table) 1)))])
     (and index (vector-ref clauses index))))

(define (Trex->AST trex)
  (define (source src ast) (AST-src-set! ast src) ast)

  (define (rec0 trex)
    (define (rec trex) (rec0 trex))

    (define (CaseLambdaClause->AST clause)
      (define (improper-list-split list default)
        (cond
         [(null? list) (values '() default)]
         [(pair? list)
          (let-values ([(front rest) (improper-list-split (cdr list) default)])
            (values (cons (car list) front) rest))]
         [else (values '() list)]))
      (Trex:Trex-CaseLambdaClause-case clause
        [(Trex:CaseLambdaClause formals body)
         (let-values ([(front rest) (improper-list-split formals #f)])
           (make-case-lambda-clause (list->vector front) rest (rec body)))]))

    (Trex:Trex-case trex
      [(Trex:Ref src id) (source src (Ref id))]
      [(Trex:Quote src obj) (source src (Quote obj))]
      [(Trex:Primref src safe name) (source src (Primref safe name))]
      [(Trex:Set src id expr) (source src (Set id (rec expr)))]
      [(Trex:App src expr expr*)
       (source src (App (rec expr) (list->vector (map rec expr*))))]
      [(Trex:If src test thn els)
       (source src (If (rec test) (rec thn) (rec els)))]
      [(Trex:CaseLambda src clause*)
       (let ([clause* (list->vector (map CaseLambdaClause->AST clause*))])
         (source src (CaseLambda (make-case-lambda-table clause*) clause*)))]
      [(Trex:Seq src expr1 expr2)
       (source src (Seq (rec expr1) (rec expr2)))]
      [(Trex:Letrec src lhs rhs body)
       (source src (Letrec
                    (list->vector (map make-letrec-binding lhs (map rec rhs)))
                    (rec body)))]))

  (define (rec1 ast context)
    (let ([context (myers-stack-cons ast context)])
      (define (rec ast) (rec1 ast context))
      (AST-context-set! ast context)

      (AST-case ast
        [(Ref var) (values)]
        [(Quote obj) (values)]
        [(Primref safe name) (values)]
        [(Set var expr) (rec expr)]
        [(App fun actuals) (rec fun) (vector-for-each rec actuals)]
        [(If test conseq altern) (rec test) (rec conseq) (rec altern)]
        [(CaseLambda table clauses)
         (vector-for-each (lambda (clause) (rec (case-lambda-clause-body clause)))
                          clauses)]
        [(Seq expr1 expr2) (rec expr1) (rec expr2)]

       [(Letrec bindings body)
        (vector-for-each (lambda (binding) (rec (letrec-binding-rhs binding)))
                         bindings)
        (rec body)])))

  (define ast (rec0 trex))
  (rec1 ast myers-stack-null)
  ast)

)
