(library (composition-stacks)

  ;; PROPERTIES:
  ;;  All operations are O(lg n) or better
  ;;   (except list->composition-stack and composition-stack->list).
  ;;
  ;;  All operations are observationally equivalent to the corresponding
  ;;   list operations.  This includes pointer equality and mutation.

  ;; NOTATION:
  ;;  obj - any scheme object
  ;;  stack - myers-stack object
  ;;  stack-pair - myers-stack object that is not null
  ;;  elem - object contained in the myers-stack
  ;;  list - list of elem objects
  ;;  length - non-negative integer less then or equal to length
  ;;  index - non-negative integer strictly less than length

  (export
   composition-stack? ;; obj -> boolean

   make-composition-stack-null ;; (elem * elem -> elem) -> stack -- associative
   composition-stack-null? ;; obj -> boolean

   composition-stack-cons ;; elem * stack -> stack
   composition-stack-pair? ;; obj -> boolean
   composition-stack-car ;; stack-pair -> obj
   ;; composition-stack-car-set! ;; stack-pair * obj -> void
   composition-stack-cdr ;; stack-pair -> stack

   composition-stack-length ;; stack -> length
   composition-stack-ref ;; stack * index -> elem
   ;; composition-stack-set! ;; stack * index * elem -> void
   list->composition-stack ;; list -> stack
   composition-stack->list ;; stack -> list

   composition-stack-find ;; (stack ... -> bool) * stack ... -> stack
                          ;; returns first that is true
                          ;; stacks must be same length
  composition-stack-suffix ;; stack * length -> stack
  composition-stack-drop ;; stack * length -> stack

  composition-stack-previous ;; stack * stack -> stack (second is suffix)
  composition-stack-previous-ref ;; stack * stack -> elem (second is suffix)
  composition-stack-eq-suffix ;; stack * stack -> stack
                              ;; returns first that is 'eq?'
                              ;; need not be same length
  composition-stack-range ;; elem * stack * stack -> elem
  )

 (import (rnrs)
         (rename
          (myers-stacks)
          (myers-stack-null? composition-stack-null?)
          (myers-stack-pair? composition-stack-pair?)
          (myers-stack-car composition-stack-car)
          ;; myers-stack-car-set! is not exported (b/c it breaks the compositions)
          (myers-stack-cdr composition-stack-cdr)
          (myers-stack-length composition-stack-length)
          (myers-stack-ref composition-stack-ref)
          (myers-stack->list composition-stack->list)
          (myers-stack-find composition-stack-find)
          (myers-stack-suffix composition-stack-suffix)
          (myers-stack-drop composition-stack-drop)
          (myers-stack-previous composition-stack-previous)
          (myers-stack-previous-ref composition-stack-previous-ref)
          (myers-stack-eq-suffix composition-stack-eq-suffix)))

;; TODO:
;;  - make car non-mutable
;;  - calculate constant for composition-stack-jump

;;;;;;;;;;;;;;;;
;; Datatype
;;;;;;;;;;;;;;;;

(define-record-type composition-stack
  (nongenerative)
  (fields composer composite)
  (parent myers-stack))

(define composition-stack-jump
  (let ([rtd (record-type-descriptor myers-stack)]
        [index 3])
    (assert (eq? 'jump (vector-ref (record-type-field-names rtd) index)))
    (record-accessor rtd index)))

;;;;;;;;;;;;;;;;
;; Operations
;;;;;;;;;;;;;;;;
(define (make-composition-stack-null composer)
  (let* ([jump-jump (make-composition-stack #f -1 #f #f #f #f)]
         [jump (make-composition-stack #f 0 #f jump-jump #f #f)])
    (make-composition-stack #f 0 #f jump composer #f)))

(define (composition-stack-cons car cdr)
  (let* ([jump (composition-stack-jump cdr)]
         [jump-jump (composition-stack-jump jump)])
    (let ([cdr-length (composition-stack-length cdr)]
          [jump-length (composition-stack-length jump)]
          [jump-jump-length (composition-stack-length jump-jump)])
      (let ([length (fx+ 1 cdr-length)]
            [composer (composition-stack-composer cdr)])
        (if (eq? (fx- cdr-length jump-length)
                 (fx- jump-length jump-jump-length))
              (make-composition-stack
               car length cdr jump-jump composer
               (composer car (composer (composition-stack-composite cdr)
                                       (composition-stack-composite jump))))
              (make-composition-stack
               car length cdr cdr composer car))))))

(define (list->composition-stack compose list)
  (fold-right
   composition-stack-cons (make-composition-stack-null compose) list))

(define (composition-stack-range identity stack suffix)
  (define (rec identity composer stack length)
    (if (fx=? length (composition-stack-length stack))
        identity
        (let ([jump (composition-stack-jump stack)])
          (if (fx>? length (composition-stack-length jump))
              (composer
               (composition-stack-car stack)
               (rec identity composer (composition-stack-cdr stack) length))
              (composer
               (composition-stack-composite stack)
               (rec identity composer jump length))))))
  (rec identity (composition-stack-composer stack)
       stack (composition-stack-length suffix)))

)
