(library (datatype)
  (export
   define-datatype ;; Macro (see comment for syntax)
   list-of ;; (obj -> bool) -> (obj -> bool)
   vector-of ;; (obj -> bool) -> (obj -> bool)
   improper-list-of ;; (obj -> bool) -> (obj -> bool)
   )
  (import (rnrs) (syntax-helpers) (records-io)
          (only (chezscheme) optimize-level void))

;; Disables checks at Chez Scheme optimize-level 3
(define-syntax unless-unchecked
  (lambda (x)
    (syntax-case x ()
      [(_ body ...)
       (if (fx=? (optimize-level) 3)
           #'(void)
           #'(begin body ...))])))

(define (list-of pred)
  (lambda (x*) (and (list? x*) (for-all pred x*))))

(define (vector-of pred)
  (lambda (x*) (and (vector? x*) (for-all pred (vector->list x*)))))

(define (improper-list-of pred)
  (lambda (x*)
    (or (null? x*)
        (pred x*)
        (and (pair? x*) (pred (car x*)) ((improper-list-of pred) (cdr x*))))))

(define-syntax define-case
  #;(syntax-summary
     (define-case name extract-tag
       [name tag extractor extractor]
       [name tag extractor extractor])

     (name-case expr
       [(name field field) body]
       [(name field field) body]
       [else body]))

  (lambda (stx)
    (syntax-case stx ()
      [(_ case-name tag-extractor [name tag field-extractor ...] ...)
       (syntax-for-all identifier? #'(name ...))
       (with-syntax ([((field ...) ...)
                      (syntax-map generate-temporaries
                                  #'((field-extractor ...) ...))])
         #'(define-syntax case-name
             (lambda (stx)
               (define (make-clause clause)
                 (syntax-case clause (name ...)
                   [[(name field ...) expr0 expr* (... ...)]
                    #'[(tag)
                       (let ([field (field-extractor expr-var)] ...)
                         #f ;; force non-definition context like scheme 'case'
                         expr0 expr* (... ...))]] ...))
               (syntax-case stx (else)
                 [(_ expr
                     [(clause-name field-name (... ...)) body0 body* (... ...)]
                     (... ...))
                  (syntax-for-all (lambda (x) (syntax-for-all identifier? x))
                                  #'((field-name (... ...)) (... ...)))
                  ;; TODO: check for missing clause
                  #'(case-name expr
                      [(clause-name field-name (... ...)) body0 body* (... ...)]
                      (... ...)
                      [else (assertion-violation 'case-name "missing case")])]
                 [(_ expr
                     [(clause-name field-name (... ...)) body0 body* (... ...)]
                     (... ...)
                     [else else-body0 else-body* (... ...)])
                  (syntax-for-all (lambda (x) (syntax-for-all identifier? x))
                                  #'((field-name (... ...)) (... ...)))
                  ;; TODO: check for duplicate clause
                  #`(let ([expr-var expr])
                      (case (tag-extractor expr-var)
                        #,@(syntax-map
                            make-clause
                            #'([(clause-name field-name (... ...))
                                body0 body* (... ...)] (... ...)))
                        [else else-body0 else-body* (... ...)]))]))))])))

(define-syntax define-datatype
  #;(syntax-summary
     (define-datatype (AST (property default type) (property default))
       [Constor (field type) field]
       [Constor field (field type)]))

  (lambda (stx)
    (define (normalize-property-spec stx)
      (syntax-case stx ()
        [(property default) (identifier? #'property)
         #'(property default (lambda (x) #t))]
        [(property default predicate) (identifier? #'property)
         #'(property default predicate)]))
    (define (normalize-field-spec stx)
      (syntax-case stx ()
        [field (identifier? #'field) #'(field (lambda (x) #t))]
        [(field predicate) (identifier? #'field)  #'(field predicate)]))
    (syntax-case stx ()
      [(_ (datatype property-spec ...)
          [constructor field-spec ...] ...)
       ;; exports:
       ;;  datatype(record) datatype?
       ;;  datatype-property datatype-property-set!
       ;;  datatype-case constructor constructor?
       (and (identifier? #'datatype) (for-all identifier? #'(constructor ...)))

       ;; expand property-spec and field-spec
       (with-syntax ([((property property-default property-predicate) ...)
                      (syntax-map normalize-property-spec #'(property-spec ...))]
                     [(((field field-predicate) ...) ...)
                      (syntax-map
                       (lambda (x) (syntax-map normalize-field-spec x))
                       #'((field-spec ...) ...))])

         ;; public exports
         (with-syntax ([datatype? (syntax-append #'datatype #'datatype "?")]
                       [datatype-case (syntax-append #'datatype #'datatype "-case")]
                       [(datatype-property ...)
                        (syntax-map
                         (lambda (x) (syntax-append #'datatype #'datatype "-" x))
                         #'(property ...))]
                       [(datatype-property-set! ...)
                        (syntax-map
                         (lambda (x)
                           (syntax-append #'datatype #'datatype "-" x "-set!"))
                         #'(property ...))])

            ;; private identifiers
            (with-syntax ([(datatype-property-unchecked-set! ...)
                           (syntax-map
                            (lambda (x)
                              (syntax-append
                               #'private #'datatype "-" x "-unchecked-set!"))
                            #'(property ...))]
                          [((constructor-field ...) ...)
                           (syntax-map
                            (lambda (constructor fields)
                              (syntax-map
                               (lambda (field)
                                 (syntax-append #'private constructor "-" field))
                               fields))
                            #'(constructor ...)
                            #'((field ...) ...))]
                          [(constructor-record ...)
                           (syntax-map (lambda (x)
                                         (syntax-append #'private x "-record"))
                                       #'(constructor ...))]
                          [(constructor? ...)
                           (syntax-map (lambda (x) (syntax-append #'private x "?"))
                                       #'(constructor ...))]
                          [(constructor-number ...)
                           (syntax-indexes #'(constructor ...))])

              ;; resulting code
              #'(begin
                  ;; Define parent record
                  ;;   private: make-datatype and datatype-variant
                  (define-record-type (datatype make-datatype datatype?)
                    (fields (immutable variant datatype-variant)
                            (mutable property datatype-property
                                     datatype-property-unchecked-set!) ...)
                    (nongenerative)
                    (protocol (lambda (make)
                                (lambda (variant)
                                  (make variant property-default ...)))))
                  (define (datatype-property-set! obj value)
                    (unless-unchecked
                     (unless (property-predicate value)
                       (assertion-violation 'datatype-property-set!
                                            "invalid value" value)))
                    (datatype-property-unchecked-set! obj value))
                  ...

                  ;; Define child records
                  (define-record-type
                    (constructor-record constructor constructor?)
                    (parent datatype)
                    (fields (immutable field constructor-field) ...)
                    (sealed #t)
                    (nongenerative)
                    (protocol
                     (lambda (parent)
                       (lambda (field ...)
                         (unless-unchecked
                          (define who 'constructor)
                          (unless (field-predicate field)
                            (assertion-violation
                             who "invalid value for field" 'field field))
                          ...)
                         ((parent constructor-number) field ...)))))
                  ...

                  ;; Define case form
                  (define-case datatype-case datatype-variant
                    [constructor constructor-number constructor-field ...] ...)

                  ;; Define reader/writer
                  (define *ignored* ;; maintain definition context
                    (begin
                      (record-reader
                       'constructor (record-type-descriptor constructor-record))
                      ...
                      (record-writer (record-type-descriptor constructor-record)
                        (lambda (x p wr)
                          (wr `(constructor
                                ,@(map (lambda (a) (a x))
                                       (list constructor-field ...))) p)))
                      ...))
                  ))))])))
)
