(library (flow-graphs)
  (export

   make-flow-graph ;; void -> flow-graph
   flow-graph? ;; obj -> bool
   flow-graph-step! ;; flow-graph -> void
   flow-graph-stable? ;; flow-graph -> bool
   flow-graph-print
   ;; (printer :
   ;;   (contain : string * list of flow-graph-node * list of obj -> void) *
   ;;   (node : obj) -> void) *
   ;; (labeler : flow-graph-node -> string) *
   ;; obj -> void

   flow-graph-node? ;; obj -> bool
   flow-graph-node-output ;; flow-graph-node -> obj
   flow-graph-node-info ;; flow-graph-node -> obj

   ;; flow-graph * info * eqv? * output -> flow-graph-input-node
   make-flow-graph-mutable-node

   ;; flow-graph-mutable-node? ;; obj -> bool
   flow-graph-mutable-node-set! ;; flow-graph-input-node * output -> void

   ;; flow-graph * info * eqv? * function * output -> flow-graph-lazy-node
   make-flow-graph-lazy-node

   ;; flow-graph-lazy-node? ;; obj -> bool
   flow-graph-lazy-node-force! ;; flow-graph-lazy-node * args ... -> void

   ;; flow-graph * info * eqv? * function * args ... -> flow-graph-function-node
   make-flow-graph-function-node

   ;; flow-graph-function-node? ;; obj -> bool

   ;; flow-graph * info * eqv? * input -> flow-graph-join-node
   make-flow-graph-join-node

   ;;flow-graph-join-node? ;; obj -> bool
   )
  (import (except (rnrs) case cond if) (safe-forms) (records-io)
          (rnrs mutable-pairs))

;;;;;;;;;;;;;;;;
;; Flow Graphs
;;;;;;;;;;;;;;;;

(define-record-type flow-graph
  (nongenerative)
  (fields (mutable pending-list)) ;; TODO: as a rotating vector of lists
  (protocol (lambda (make) (lambda () (make '())))))

(define (flow-graph-pending-add! flow-graph node)
  (flow-graph-pending-list-set!
   flow-graph (cons node (flow-graph-pending-list flow-graph))))

(define (flow-graph-pending-remove! flow-graph)
  (let ([x (flow-graph-pending-list flow-graph)])
    (flow-graph-pending-list-set! flow-graph (cdr x))
    (car x)))

(define (flow-graph-step! graph)
  (define node (flow-graph-pending-remove! graph))
  (when (flow-graph-node-pending node)
        (flow-graph-node-pending-set! node #f)
        (node-output-set! node ((flow-graph-node-action node) node))))

(define (flow-graph-stable? graph) (null? (flow-graph-pending-list graph)))

;;;;;;;;;;;;;;;;
;; Graph Nodes
;;;;;;;;;;;;;;;;
(define-record-type flow-graph-node
  (nongenerative)
  (fields graph
          info
          (mutable eqv?)
          (mutable pending) ;; #t or #f
          (mutable output)
          #;(mutable outputs) ;; for debugging
          (mutable action) ;; void -> output
          (mutable listeners) ;; list of (node or (cons (node or #f) '()))
          )
  (protocol
   (lambda (make)
     (lambda (graph info eqv? output action)
       (assert (string? info))
       (assert (or (not eqv?) (eqv? output output))) ;; Sanity check
       (make graph info eqv? #f output action '())))))

;;;;;;;;;;;;;;;;
;; Lazy Node
;;;;;;;;;;;;;;;;
(define (flow-graph-lazy-node-action node)
  (error 'flow-graph-lazy-node "evaluating unforced lazy node" node))
(define (make-flow-graph-lazy-node graph info eqv? output)
  (make-flow-graph-node graph info eqv? output flow-graph-lazy-node-action))

(define-syntax flow-graph-lazy-node-force!
  (lambda (stx)
    (syntax-case stx ()
      [(_ n fun args ...)
       (with-syntax ([(arguments ...) (generate-temporaries #'(args ...))])
         #'(let ([node n]
                 [function fun]
                 [arguments args] ...)
             (assert (eq? (flow-graph-node-graph node)
                          (flow-graph-node-graph arguments))) ...
             (assert (eq? (flow-graph-node-action node)
                          flow-graph-lazy-node-action))
             (flow-graph-node-action-set!
              node (lambda (node)
                     (function (flow-graph-node-output arguments) ...)))
             (add-to-listeners! arguments node) ...
             (add-to-pending! node)))])))

;;;;;;;;;;;;;;;;
;; Function Node
;;;;;;;;;;;;;;;;
(define-syntax make-flow-graph-function-node
  (lambda (stx)
    (syntax-case stx ()
      [(_ graph info eqv? function arguments ...)
       (with-syntax ([(fun args ...)
                      (generate-temporaries #'(function arguments ...))])
         #'(let ([g graph]
                 [fun function]
                 [args arguments] ...)
             (assert (eq? g (flow-graph-node-graph args))) ...
             (let ([action (lambda (node) (fun (flow-graph-node-output args) ...))])
               (let ([node (make-flow-graph-node g info eqv? (action #f) action)])
                 (add-to-listeners! args node) ...
                 node))))])))

;;;;;;;;;;;;;;;;
;; Mutable Node
;;;;;;;;;;;;;;;;
(define (flow-graph-mutable-node-action node)
  (error 'flow-graph-mutable-node "shouldn't be on pending queue" node))
(define (make-flow-graph-mutable-node graph info eqv? output)
  (make-flow-graph-node graph info eqv? output flow-graph-mutable-node-action))
(define (flow-graph-mutable-node? node)
  (and (flow-graph-node? node)
       (eq? (flow-graph-node-action node) flow-graph-mutable-node-action)))

(define (flow-graph-mutable-node-set! node new-output)
  (assert (eq? (flow-graph-node-action node) flow-graph-mutable-node-action))
  (node-output-set! node new-output))

;;;;;;;;;;;;;;;;
;; Join Node
;;;;;;;;;;;;;;;;
(define-syntax make-flow-graph-join-node
  (lambda (stx)
    (syntax-case stx ()
      [(_ graph info eqv? function arguments ...)
       (with-syntax ([(fun args ...)
                      (generate-temporaries #'(function arguments ...))])
         #'(let ([g graph]
                 [fun function]
                 [args arguments] ...)
             (assert (eq? g (flow-graph-node-graph args))) ...
             (let ([inner-node #f] ;; dummy value
                   [key (cons #f '())]) ;; dummy value
               (let ([action
                      (lambda (node)
                        (let ([new-inner (fun (flow-graph-node-output args) ...)])
                          ;; (assert (eq? graph (flow-graph-node-graph inner-node)))
                          (when (not (eq? inner-node new-inner))
                            (remove-from-join-listeners! key)
                            (set! inner-node new-inner)
                            (set! key (add-to-join-listeners! new-inner node)))
                          (flow-graph-node-output inner-node)))])
                 (let ([node (make-flow-graph-node graph info
                                                   #f #f ;; dummy values
                                                   action)])
                   (add-to-listeners! args node) ...
                   (let ([output (action node)])
                     (assert (eqv? output output)) ;; ensure dummy values skipped
                     (flow-graph-node-eqv?-set! node eqv?) ;; fix dummy value
                     (flow-graph-node-output-set! node output)) ;; fix dummy value
                   node)))))])))

;;;;;;;;;;;;;;;;
;; Printing
;;;;;;;;;;;;;;;;
;; (printer :
;;   (contain : string * list of flow-graph-node * list of obj -> void) *
;;   (node : obj) -> void) *
;; (labeler : flow-graph-node -> string) *
;; obj -> void

(define (flow-graph-print container node-label node)
  ;; for indentation
  (define depth 0)
  (define (indent)
    (let loop ([i depth]) (unless (zero? i) (display "  ") (loop (- i 1)))))

  ;; for labeling nodes
  (define node-table (make-eq-hashtable))
  (define (node-table-add-node! node)
    (assert (not (hashtable-contains? node-table node)))
    (hashtable-set! node-table node (cons containers (hashtable-size node-table))))
  (define (node-id node)
    (cdr (or (hashtable-ref node-table node #f)
             (assertion-violation 'node-id "node doesn't exist" node))))
  (define (node-cluster node)
    (car (or (hashtable-ref node-table node #f)
             (assertion-violation 'node-container "node doesn't exist" node))))
  (define (node-name node) (string-append "node" (number->string (node-id node))))

  (define containers 0) ;; for labeling clusters

  ;; for printing missing edges
  (define node-container-table (make-eq-hashtable))

  (define (escape-label x) (string-append "\"" x "\""))
  (define edges '())

  (define (print-containment label nodes children)
    (indent)(display "subgraph cluster")(display containers)(display " {\n")
    (set! depth (+ depth 1))
    (set! containers (+ containers 1))
    (indent)(display "label = ")(display (escape-label label))(display ";\n")
    (for-each (lambda (child) (container print-containment child)) children)
    (for-each (lambda (x) (print-node x #t)) nodes)
    (set! depth (- depth 1))
    (indent)(display "}\n"))

  (define (print-node node first-pass)
    (assert (flow-graph-node? node))
    (node-table-add-node! node)
    (when (and first-pass #t #;(flow-graph-lazy-node? node))
      (set! edges
            (append (map (lambda (x) (cons node x))
                         (filter (lambda (x) (not (pair? x)))
                                 (flow-graph-node-listeners node)))
                    edges)))
    (indent)
    (unless first-pass
      (display "subgraph cluster")(display (node-cluster node))(display "{ "))
    (display (node-name node))
    (display " [")
    (display "label = ")(display (escape-label (node-label node)))
    (display ", shape = ")
    (display "ellipse")
    #;(cond
        [(flow-graph-mutable-node? node) (display "diamond")]
        [(flow-graph-lazy-node? node) (display "ellipse")])
    (when (flow-graph-node-pending node)
      (display ", penwidth = 4"))
    (unless first-pass (display ", style=dotted"))
    (display "];")
    (unless first-pass (display "}"))
    (display "\n"))

  (display "digraph graph0 {\n")
  (container print-containment node)
  (for-each (lambda (edge)
              (define from (car edge))
              (define to (cdr edge))
              (unless (hashtable-contains? node-table from) (print-node from #f))
              (display (node-name from))(display " -> ")
              (display (node-name to))(display ";\n"))
            edges)
  (display "}\n"))

;;;;;;;;;;;;;;;;
;; Private Functions
;;;;;;;;;;;;;;;;

(define dummy-pair (cons #f #f))
(define (node-output-set! node new-output)
  (let ([old-output (flow-graph-node-output node)])
    (when (not ((flow-graph-node-eqv? node) old-output new-output))
      #;(flow-graph-node-outputs-set!
          node (cons old-output (flow-graph-node-outputs node)))
      (flow-graph-node-output-set! node new-output)
      (let loop ([list (flow-graph-node-listeners node)]
                 [prev dummy-pair]) ;; just to get the loop started
        (if (null? list)
            (values)
            (let ([x (car list)]
                  [rest (cdr list)])
              (if (not (pair? x))
                  (add-to-pending! x)
                  (let ([car-x (car x)])
                    (if car-x
                        (add-to-pending! car-x)
                        (set-cdr! prev rest))))
              (loop rest list))))
      (set-cdr! dummy-pair #f))))

(define (add-to-pending! node)
  (assert (not (eq? (flow-graph-node-action node) flow-graph-mutable-node-action)))
  (flow-graph-node-pending-set! node #t)
  (flow-graph-pending-add! (flow-graph-node-graph node) node))

(define (add-to-listeners! node listener)
  (flow-graph-node-listeners-set!
   node (cons listener (flow-graph-node-listeners node))))

(define (add-to-join-listeners! node listener)
  (let ([key (list listener)])
    (add-to-listeners! node key)
    key))

(define (remove-from-join-listeners! key)
  (set-car! key #f))

(record-writer (record-type-descriptor flow-graph-node)
  (lambda (r p wr)
    (display "#[node " p)
    (display (flow-graph-node-info r) p)
    (display " " p)
    (display (flow-graph-node-output r) p)
    (display "]" p)))

)
