(library (myers-stacks)

  ;; PROPERTIES:
  ;;  All operations are O(lg n) or better
  ;;   (except list->myers-stack and myers-stack->list).
  ;;
  ;;  All operations are observationally equivalent to the corresponding
  ;;   list operations.  This includes pointer equality and mutation.

  ;; NOTATION:
  ;;  obj - any scheme object
  ;;  stack - myers-stack object
  ;;  stack-pair - myers-stack object that is not null
  ;;  elem - object contained in the myers-stack
  ;;  list - list of elem objects
  ;;  length - non-negative integer less then or equal to length
  ;;  index - non-negative integer strictly less than length

  (export
   myers-stack
   myers-stack? ;; obj -> boolean

   myers-stack-null ;; stack
   myers-stack-null? ;; obj -> boolean

   myers-stack-cons ;; elem * stack -> stack
   myers-stack-pair? ;; obj -> boolean
   myers-stack-car ;; stack-pair -> elem
   myers-stack-car-set! ;; stack-pair * elem -> void
   myers-stack-cdr ;; stack-pair -> stack

   myers-stack-length ;; stack -> length
   myers-stack-ref ;; stack * index -> elem
   list->myers-stack ;; list -> stack
   myers-stack->list ;; stack -> list

   myers-stack-find ;; (stack ... -> bool) * stack ... -> stack
                    ;; returns first that is true
                    ;; stacks must be same length
   myers-stack-suffix ;; stack * length -> stack
   myers-stack-drop ;; stack * length -> stack

   myers-stack-previous ;; stack * stack -> stack (second is suffix)
   myers-stack-previous-ref ;; stack * stack -> elem (second is suffix)
   myers-stack-eq-suffix ;; stack * stack -> stack
                         ;; returns first that is 'eq?'
                         ;; stacks need not be same length
   )

  (import (except (rnrs) case cond if) (safe-forms) (records-io))

;; TODO:
;;  - use fx or not?
;;  - fold-right in O(lg n)
;;  - myers-stack-find is only valid if args are same length
;;     (maybe call it find-from-rear and trim arguments to same length)

;;;;;;;;;;;;;;;;
;; Datatype
;;;;;;;;;;;;;;;;

;; Defines the following exported functions:
;;  myers-stack?
;;  myers-stack-car
;;  myers-stack-car-set!
;;  myers-stack-cdr
;;  myers-stack-length
(define-record-type myers-stack
  (nongenerative)
  (fields (mutable $car) length $cdr jump))

(define myers-stack-car-set! myers-stack-$car-set!)

(define (myers-stack-car stack)
  (assert (< 0 (myers-stack-length stack)))
  (myers-stack-$car stack))

(define (myers-stack-cdr stack)
  (assert (< 0(myers-stack-length stack)))
  (myers-stack-$cdr stack))

;;;;;;;;;;;;;;;;
;; Basic Ops
;;;;;;;;;;;;;;;;
(define myers-stack-null
  (let* ([jump-jump (make-myers-stack #f -1 #f #f)]
         [jump (make-myers-stack #f 0 #f jump-jump)])
    (make-myers-stack #f 0 #f jump)))

(define (myers-stack-null? stack)
  (and (myers-stack? stack) (fxzero? (myers-stack-length stack))))

(define (myers-stack-cons car cdr)
  (let* ([jump (myers-stack-jump cdr)]
         [jump-jump (myers-stack-jump jump)])
    (let ([cdr-length (myers-stack-length cdr)]
          [jump-length (myers-stack-length jump)]
          [jump-jump-length (myers-stack-length jump-jump)])
      (let ([length (fx+ 1 cdr-length)])
        (if (eq? (fx- cdr-length jump-length)
                 (fx- jump-length jump-jump-length))
            (make-myers-stack car length cdr jump-jump)
            (make-myers-stack car length cdr cdr))))))

(define (myers-stack-pair? stack)
  (and (myers-stack? stack) (not (fxzero? (myers-stack-length stack)))))

(define (list->myers-stack list)
  (fold-right myers-stack-cons myers-stack-null list))

(define (myers-stack->list stack)
  (cond
    [(myers-stack-null? stack) '()]
    [(myers-stack-pair? stack)
     (cons (myers-stack-car stack)
           (myers-stack->list (myers-stack-cdr stack)))]))

(define (myers-stack-ref stack index)
  ($check-index 'myers-stack-ref stack index)
  (myers-stack-car (myers-stack-drop stack (- index 1))))

;;;;;;;;;;;;;;;;
;; Basic Finds
;;;;;;;;;;;;;;;;

(define ($check-range who stack index assertion)
  (unless assertion
    (assertion-violation who "out of range" index stack)))
(define ($check-index who stack index)
  ($check-range who stack index (< index (myers-stack-length stack))))
(define ($check-length who stack length)
  ($check-range who stack length (<= length (myers-stack-length stack))))

;; TODO: what to do if nothing matches the find?
;; TODO: rename to find-first
;; TODO: returns #f if nothing matches
(define (myers-stack-find f . args)
  (cond
    [(apply f args) (apply values args)]
    [(myers-stack-null? (car args)) #f]
    [else
     (let ([jumps (map myers-stack-jump args)])
       (cond
         ;; jump is too far
         [(apply f jumps)
          (apply myers-stack-find f (map myers-stack-cdr args))]
         ;; jump isn't too far
         [else (apply myers-stack-find f jumps)]))]))

(define-syntax myers-stack-find-macro
  (lambda (stx)
    (syntax-case stx ()
      [(_ fun x0 xs ...)
       (with-syntax ([(arg0 args ...) (generate-temporaries #'(x0 xs ...))]
                     [(jump0 jumps ...) (generate-temporaries #'(x0 xs ...))])
         #'(let loop ([f fun][arg0 x0][args xs] ...)
             (cond
               [(f arg0 args ...) (values arg0 args ...)]
               [(myers-stack-null? arg0) #f]
               [else
                (let ([jump0 (myers-stack-jump arg0)]
                      [jumps (myers-stack-jump args)] ...)
                  (cond
                    [(f jump0 jumps ...) (loop f (myers-stack-cdr arg0) (myers-stack-cdr args) ...)]
                    [else (loop f jump0 jumps ...)]))])))])))

(define (myers-stack-suffix stack length)
  ($check-length 'myers-stack-suffix stack length)
  (myers-stack-find-macro
   (lambda (stack) (<= (myers-stack-length stack) length)) stack))

(define (myers-stack-drop stack length)
  ($check-length 'myers-stack-drop stack length)
  (myers-stack-suffix stack (- (myers-stack-length stack) length)))

;;;;;;;;;;;;;;;;
;; Fancy Finds
;;;;;;;;;;;;;;;;

(define (myers-stack-previous stack suffix)
  (myers-stack-suffix stack (+ (myers-stack-length suffix) 1)))

(define (myers-stack-previous-ref stack suffix)
  (and (< (myers-stack-length stack) (myers-stack-length suffix))
       (myers-stack-car (myers-stack-previous stack suffix))))

(define (myers-stack-eq-suffix arg0 arg1)
  (let ([length (min (myers-stack-length arg0)
                     (myers-stack-length arg1))])
    (let-values ([(common0 common1)
                  (myers-stack-find-macro
                   eq?
                   (myers-stack-suffix arg0 length)
                   (myers-stack-suffix arg1 length))])
        common0)))

;;;;;;;;;;;;;;;;
;; Display
;;;;;;;;;;;;;;;;
(record-writer
  (record-type-descriptor myers-stack)
  (lambda (r p wr)
    (display "#[" p)
    (display (record-type-name (record-rtd r)) p)
    (display " " p)
    (wr (myers-stack->list r) p)
    (display "]" p)))

)
