(library (parse-scheme)
  (export parse-scheme parse-unsafe-scheme unparse-scheme)
  (import (rnrs) (tprogram) (match) (idents))

;; This module parses s-expressions into Trex records.  It is for demonstration
;; purposes only and should not be considered a robust scheme parser.

;; unique-name produces a unique name derived the input name by
;; adding a unique suffix of the form .<digit>+.
(define unique-name
  (let ([count 0])
    (lambda (sym)
      (set! count (+ count 1))
      (string->symbol
       (string-append (symbol->string sym) "." (number->string count))))))

;; improper-map -- helper function for getting a list of unique variables
;;                 when we get imporper formals lists.
(define (improper-map f ls . more)
  (if (null? more)
      (let map1 ((ls ls))
        (cond
         [(null? ls) '()]
         [(pair? ls) (cons (f (car ls)) (map1 (cdr ls)))]
         [else (f ls)]))
      (let map-more ((ls ls) (more more))
        (cond
         [(null? ls) '()]
         [(pair? ls) (cons (apply f (car ls) (map car more))
                           (map-more (cdr ls) (map cdr more)))]
         [else (apply f ls more)]))))

(define (improper-memq s ls)
  (cond
   [(null? ls) #f]
   [(and (pair? ls) (eq? s (car ls))) (cdr ls)]
   [(pair? ls) (improper-memq s (cdr ls))]
   [else (and (eq? s ls) '())]))

(define who 'parse-scheme)

;; parse-scheme accepts a single value, verifies that the value
;; is a valid program in the source language, replaces bound variables
;; with unique variables, and simplifies several kinds of expressions.
;;
;; The grammar changes from Assignment 14 in that it allows unquoted
;; fixnum, boolean, and empty-list constants; variables are arbitrary
;; symbols; lambda, let, and letrec bodies are implicit begin expressions;
;; one-armed if expressions are allowed; unquoted fixnums and booleans
;; are allowed; quoted vector/list structure is allowed; "and" expressions
;; are allowed; and "or" expressions are allowed.  In addition, "not" is
;; included in the set of primitives.
;;
;;  Program --> <Expr>
;;  Expr    --> <prim>
;;           |  <constant>
;;           |  <var>
;;           |  (quote <datum>)
;;           |  (if <Expr> <Expr>)
;;           |  (if <Expr> <Expr> <Expr>)
;;           |  (and <Expr>*)
;;           |  (or <Expr>*)
;;           |  (begin <Expr>* <Expr>)
;;           |  (lambda (<var>*) <Expr>+)
;;           |  (let ([<var> <Expr>]*) <Expr>+)
;;           |  (letrec ([<var> <Expr>]*) <Expr>+)
;;           |  (set! <var> <Expr>)
;;           |  (<Expr> <Expr>*)
;;
;; Where constant is #t, #f, (), a symbol or an exact integer
;;       var is an arbitrary symbol
;;       datum is a constant, pair of datums, or vector of datums
;;       primitives are void, error, read (zero arguments); car, cdr, 
;;         vector-length, make-vector, not, boolean?, integer?, bignum?,
;;         fixnum?, null?, pair?, procedure?, symbol?, vector?
;;         (one argument); *, +, -, /, <, <=, =, >=, >, fx*, fx+, fx-,
;;         fx/, fx<, fx<=, fx=, fx>=, fx>, cons, vector-ref, eq?, 
;;         set-car!, set-cdr! (two arguments); and vector-set!
;;         (three arguments).
;;
;; A var may not be bound multiple times by the same lambda, let, or
;; letrec expression, but a var may be bound by multiple lambda, let,
;; and letrec expressions, with its scope determined via the usual
;; Scheme rules.  No names are reserved, so a keyword or primitive name
;; loses its special meaning within a the scope of a lambda, let, or
;; letrec binding binding of the name.

(define (make-seq expr*)
  (define (strip-begin-seq expr*)
    (match `(begin ,@expr*)
      [(begin ,[expr*] ...) (apply append expr*)]
      [(seq ,[expr*] ...) (apply append expr*)]
      [,expr (list expr)]))
  (define (make-seq expr*)
    (match expr*
      [(,x) x]
      [(,x* ... ,x) (Seq #f (make-seq x*) x)]))
  (make-seq (strip-begin-seq expr*)))

;;; expr -> var
;;;         (quote src datum)
;;;         (primref src safe name)
;;;         (set! src var expr)
;;;         (app src expr expr*)
;;;         (if src test then else)
;;;         (case-lambda src clause*)
;;;         (seq src expr expr)
;;;         (letrec src ([var expr]) expr)
;;;         (type/expr src expr ty key stuff ...)  ; unconditionally evaluate expr,

(define (verify-var-list x*)
  (let loop ([x* x*])
    (cond
     [(null? x*) #f]
     [(pair? x*)
      (let ([x (car x*)] [x* (cdr x*)])
        (unless (or (symbol? x) (ident? x))
          (assertion-violation who "invalid list found" x))
        (when (improper-memq x x*)
          (assertion-violation who "non-unique var list found" x))
        (loop x*))]
     [else (unless (or (symbol? x*) (ident? x*))
             (assertion-violation who "invalid var list found" x*))])))

(define (constant? x)
  (or (memq x '(#t #f ())) (and (integer? x) (exact? x)) (string? x) (char? x)))

(define (datum? x)
  (or (constant? x)
      (symbol? x)
      (if (pair? x)
          (and (datum? (car x)) (datum? (cdr x)))
          (and (vector? x) (for-all datum? (vector->list x))))))

(define (uvar-maker sym) (unique-name sym))
(define (make-uvar-ident symbol)
  (let ([x (uvar-maker symbol)]) (make-ident x x #f)))
(define (wrap-var v) (if (ident? v) v (make-uvar-ident v)))
(define (unique-var) (make-uvar-ident 't))

(define empty-env '())
(define (extend env id x) (if (ident? id) env (cons (cons id x) env)))
(define (lookup id env) (cond [(assq id env) => cdr] [else id]))

(define (extend-variables env var* uvar*)
  (cond
   [(null? var*) env]
   [(pair? var*) (extend-variables
                  (extend env (car var*) (car uvar*))
                  (cdr var*) (cdr uvar*))]
   [else (extend env var* uvar*)]))

(define prim-list
  '(+ - * / <= < = >= > fx+ fx- fx* fx/ fx<= fx< fx= fx>= fx> $fxu<
    boolean? car cdr cons eq? integer? bignum? fixnum? symbol? make-vector
    null? pair? fxzero? procedure? set-car! set-cdr!
    vector? vector-length vector-ref vector-set!
    void error display read memq for-all))

(define (Expr env safe)
  (lambda (x)
    (define Application
      (lambda (env proc args) (App #f proc (map (Expr env safe) args))))

    (match x
      [,expr (guard (Trex? expr)) expr] ;; HACK
      [,expr (guard (ident? expr)) (Ref #f expr)] ;; HACK
      ;; bound identifiers
      [,id (guard (and (symbol? id) (ident? (lookup id env))))
           (Ref #f (lookup id env))]
      [(,id ,arg* ...) (guard (and (symbol? id)) (ident? (lookup id env)))
       (Application env (Ref #f (lookup id env)) arg*)]
      [,id (guard (and (symbol? id) (memq id prim-list)))
       (Primref #f safe id)]

      ;; special forms
      [,k (guard (constant? k)) (Quote #f k)]

      [(quote ,datum)
       (assert (datum? datum))
       (Quote #f datum)]

      [(if ,[t] ,[c] ,[a])
       (If #f t c a)]
      [(if ,[t] ,[c])
       (If #f t c (App #f (Primref #f #t 'void) (list)))]
      [(not ,[x])
       (If #f x (Quote #f #f) (Quote #f #t))]

      [(and ,[x*] ...)
       (if (null? x*)
           (Quote #f #t)
           (fold-left (lambda (a b) (If #f a b (Quote #f #f))) (car x*) (cdr x*)))]
      [(or ,[x*] ...)
       (if (null? x*)
           (Quote #f #f)
           (fold-left
            (lambda (a b)
              (let ([t (unique-var)])
                (let ([t-ref (Ref #f t)])
                  (App
                   #f (CaseLambda
                       #f (list (CaseLambdaClause (list t) (If #f t-ref t-ref b))))
                   `(,a)))))
            (car x*) (cdr x*)))]

      [(begin ,[x] ,[x*] ...)
       (make-seq `(,x ,x* ...))]

      [(lambda ,fml* ,x ,x* ...)
       (verify-var-list fml*)
       (let ([uvar* (improper-map wrap-var fml*)])
         (let ([env (extend-variables env fml* uvar*)])
           (CaseLambda
            #f (list (CaseLambdaClause
                      uvar* (make-seq (map (Expr env safe) `(,x ,x* ...))))))))]

      [(case-lambda [,fml** ,x* ,x** ...] ...)
       (for-each (lambda (fml*) (verify-var-list fml*)) fml**)
       (let ([uvar** (map (lambda (ls) (improper-map wrap-var ls)) fml**)])
         (let ([env* (map (lambda (fml* uvar*) (extend-variables env fml* uvar*))
                          fml** uvar**)])
           (let ([body* (map (lambda (x x* env)
                               (make-seq (map (Expr env safe) `(,x ,x* ...))))
                             x* x** env*)])
             (CaseLambda #f (map CaseLambdaClause uvar** body*)))))]

      [(let ([,lhs* ,[rhs*]] ...) ,x ,x* ...)
       (verify-var-list lhs*)
       (let ([uvar* (map wrap-var lhs*)])
         (let ([env (extend-variables env lhs* uvar*)])
           (App
            #f (CaseLambda
                #f (list (CaseLambdaClause
                          uvar* (make-seq (map (Expr env safe) `(,x ,x* ...))))))
            rhs*)))]

        [(letrec ([,lhs* ,rhs*] ...) ,x ,x* ...)
         (verify-var-list lhs*)
         (let ([uvar* (map wrap-var lhs*)])
           (let ([env (extend-variables env lhs* uvar*)])
             (Letrec
              #f uvar* (map (Expr env safe) rhs*)
              (make-seq `(,((Expr env safe) x) ,(map (Expr env safe) x*) ...)))))]

        [(set! ,id ,[rhs])
         (guard (symbol? id))
         (let ([uvar (lookup id env)])
           (cond
            [(ident? uvar) (ident-assigned-set! uvar #t) (Set #f uvar rhs)]
            [else (App #f (Primref #f #f '$set-top-level-value!)
                       (list (Quote #f uvar) rhs))]))]

        [(apply ,[rator] ,[rand*] ... ,[(Expr env safe) -> rest])
         (App #f (Primref #f #t 'apply) `(,rator ,rand* ... ,rest))]

        ;; application
        [(,[x] ,arg* ...) (Application env x arg*)]

        ;; free identifiers
        [,id
         (guard (symbol? id))
         (begin
           (display (list "WARNING: top-level-value" id))(newline)
           (App #f (Primref #f #f '$top-level-value) (list (Quote #f id))))]

        [,x (assertion-violation who "invalid Expr" x)])))

(define parse-scheme (Expr empty-env #t))
(define parse-unsafe-scheme (Expr empty-env #f))

(define (unparse-scheme x)
  (define (Ref->ident x)
    (Trex-case x
      [(Ref src id) id]
      [else #f]))
  (Trex-case x
    [(Ref src id) (ident-name id)]
    [(Quote src datum) `(quote ,datum)]
    [(Seq src e1 e2) `(begin ,(unparse-scheme e1) ,(unparse-scheme e2))]
    [(CaseLambda src cl*)
     (let ([cl*
            (map (lambda (cl)
                   (Trex-CaseLambdaClause-case
                    cl [(CaseLambdaClause fmls body)
                        `(,(improper-map ident-name fmls) ,(unparse-scheme body))]))
                 cl*)])
       (if (= (length cl*) 1)
           `(lambda ,(caar cl*) ,(cadar cl*))
           `(case-lambda ,@cl*)))]
    [(App src rator rand*)
     `(,(unparse-scheme rator) ,@(map unparse-scheme rand*))]
    [(Letrec src var* expr* body)
     `(letrec ,(map (lambda (var expr)
                      `[,(ident-name var) ,(unparse-scheme expr)]) var* expr*)
        ,(unparse-scheme body))]
    [(If src t c a)
     `(if ,(unparse-scheme t) ,(unparse-scheme c) ,(unparse-scheme a))]
    [(Set src var expr) `(set! ,(ident-name var) ,(unparse-scheme expr))]
    [(Primref src safe name)
     (if safe
         name
         (string->symbol (string-append "unsafe-" (symbol->string name))))]))

)
