(library (prims)
  (export prim-name->type)
  (import (except (rnrs) case cond if) (safe-forms) (idents) (types) (bitypes)
          (syntax-helpers) (ast-flow-graphs) (flow-graphs))

(define (make-AST-const-node ast name const)
  (make-AST-node ast make-flow-graph-mutable-node name eq? const))

(define-syntax make-prim-core
  (lambda (stx)
    (syntax-case stx ()
      [(_ name [(return-true return-false)
                (arg-escapes arg-true arg-false) ...
                (has-rest rest-escapes rest-true rest-false)] ...)
       (with-syntax ([(return ...)
                      (generate-temporaries #'(return-true ...))]
                     [((arg ...) ...)
                      (syntax-map generate-temporaries #'((arg-true ...) ...))]
                     [(rest ...)
                      (generate-temporaries #'(rest-true ...))]
                     [((formal-position ...) ...)
                      (syntax-map syntax-indexes #'((arg-true ...) ...))]
                     [(clause-arg-count ...)
                      (syntax-map syntax-length #'((arg-true ...) ...))]
                     )
         #'(cons
            'name
            (let ([name
                   (lambda (ast)
                     (define bot
                       (make-AST-const-node ast "no-return" bot-type))
                     (define bibot
                       (make-AST-const-node ast "no-return" (type->bitype bot-type)))
                     (define return
                       (make-AST-const-node ast "return"
                         (type-union (maybe-non-proc-type->type return-true)
                                     (maybe-non-proc-type->type return-false)))) ...
                     (define arg
                       (make-AST-const-node
                        ast "arg"
                        (make-bitype (maybe-non-proc-type->type arg-true)
                                     (maybe-non-proc-type->type arg-false)))) ... ...
                     (define rest
                       (make-AST-const-node
                        ast "rest" (make-bitype
                                    (maybe-non-proc-type->type rest-true)
                                    (maybe-non-proc-type->type rest-false)))) ...
                     (make-proc-type
                      (lambda () '()) ;; escape
                      (lambda (arg-count called) '()) ;; input-called
                      (lambda (arg-count position type) ;; input-argument
                        (when (or (and
                                   ((if has-rest >= =) arg-count clause-arg-count)
                                   (or (and arg-escapes (= position formal-position))
                                       ...
                                       (and has-rest rest-escapes
                                            (>= position clause-arg-count))))
                                  ...)
                          (mark-escaped-type! type)
                          #f)) ;; TODO: void instead of #f
                      (lambda (arg-count) ;; output-return
                        (cond
                         [((if has-rest >= =) arg-count clause-arg-count)
                          return]
                         ...
                         [else bot]))
                      (lambda (arg-count position) ;; output-argument
                        (cond
                         [((if has-rest >= =) arg-count clause-arg-count)
                          (cond
                           [(= position formal-position) arg]
                           ...
                           [(and has-rest (>= position clause-arg-count)) rest]
                           [else bibot])]
                         ...
                         [else bibot]))))])
              name)))])))

(define-syntax make-prim-rest
  (lambda (stx)
    (define (add-rest stx)
      (syntax-case stx ()
        [((return-true return-false) args ... (escapes true false) dots)
         (syntax-ellipsis? #'dots)
         #'((return-true return-false) args ... (#t escapes true false))]
        [((return-true return-false) args ... (dots true false))
         (syntax-ellipsis? #'dots)
         ;; Fix problem with make-top-prim
         (add-rest #'((return-true return-false) args ... dots))]
        [((return-true return-false) args ...)
         #'((return-true return-false) args ... (#f #f bot-type bot-type))]))
    (syntax-case stx ()
      [(_ name clauses ...)
       #`(make-prim-core name #,@(syntax-map add-rest #'(clauses ...)))])))

(define-syntax make-prim
  (syntax-rules ()
    [(_ (names ...) clauses ...) (list (make-prim-rest names clauses ...) ...)]))

(define-syntax make-top-prim
  (syntax-rules ()
    [(_ (names ...) [(true-type false-type) arg-escapes ...] ...)
     (make-prim (names ...)
                [(true-type false-type) (arg-escapes top-type top-type) ...]
                ...)]))

(define-syntax make-pred-prim
  (syntax-rules ()
    [(_ (names ...) type)
     (make-prim (names ...)
                [(true-type-bits false-type-bits) (#f type (not-type type))])]))

;; TODO: expand arithmetic (do as wrapper around make-prim?)
;;   (may be unneeded in new design b/c proc-type is used instead of case-lambda)
;;   (same logic for in argument position?)
;; NOTE: for non-smart prims argument output types are
;;   the return types assuming prim has escaped
;; TODO: case for bare vs in application

;; TODO: as a hashtable
(define prim-table
  (append

   (make-prim (foreign-procedure) [(proc-type bot-type)
                                   (#f string-type-bits top-type)])

   (make-prim (foreign-callable) [(other-type-bits ;; code object
                                   bot-type)
                                  (#t proc-type non-proc-type-bits)])

   (make-prim (error)
     [(bot-type bot-type)
      (#f symbol-type-bits symbol-type-bits)
      (#f string-type-bits string-type-bits)
      (#f top-type top-type) ...])

   ;; TODO: escape should return the type of its argument
   (make-top-prim (escape) [(top-type top-type) #t])

   (make-prim (for-all)
     [(true-type false-type-bits)
      (#t proc-type proc-type)
      (#f pair-or-null-type-bits pair-or-null-type-bits)
      (#f pair-or-null-type-bits pair-or-null-type-bits) ...])

   ;; pair related primitives
   (make-pred-prim (pair?) pair-type-bits)

   ;; on false return could still be a pair b/c of improper lists
   (make-prim (list?)
     [(true-type-bits false-type-bits)
      (#f pair-or-null-type-bits (not-type null-type-bits))])

   (make-top-prim (cons) [(pair-type-bits bot-type) #t #t])

   (make-top-prim (list) [(pair-type-bits bot-type) #t ...])

   (make-prim (car cdr) [(true-type false-type)
                         (#f pair-type-bits pair-type-bits)])

   (make-prim (for-each)
     [(top-type top-type)
      (#t proc-type proc-type)
      (#f pair-or-null-type-bits pair-or-null-type-bits)
      (#f pair-or-null-type-bits pair-or-null-type-bits) ...])

   (make-prim (map)
     [(pair-or-null-type-bits bot-type)
      (#t proc-type proc-type)
      (#f pair-or-null-type-bits pair-or-null-type-bits)
      (#f pair-or-null-type-bits pair-or-null-type-bits) ...])

   (make-prim (reverse)
     [(pair-or-null-type-bits bot-type)
      (#f pair-or-null-type-bits pair-or-null-type-bits)])

   (make-prim (list->vector)
     [(vector-type-bits bot-type)
      (#f pair-or-null-type-bits pair-or-null-type-bits)])

   (make-prim (r6rs:number->string number->string)
     [(string-type-bits bot-type)
      (#f number-type-bits number-type-bits)]
     [(string-type-bits bot-type)
      (#f number-type-bits number-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)]
     [(string-type-bits bot-type)
      (#f number-type-bits number-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)])

   (make-prim (r6rs:string->number string->number)
     [(top-type bot-type)                     ; maybe-number, no way to express it
      (#f string-type-bits string-type-bits)]
     [(top-type bot-type)
      (#f string-type-bits string-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)])

   (make-prim (char?)
     [(true-type-bits false-type-bits)
      (#f char-type-bits (not-type char-type-bits))])

   (make-prim (put-char)
     [(other-type-bits bot-type)
      (#f other-type-bits other-type-bits)
      (#f char-type-bits char-type-bits)])

   (make-prim (write-char)
     [(other-type-bits bot-type)
      (#f char-type-bits char-type-bits)]
     [(other-type-bits bot-type)
      (#f char-type-bits char-type-bits)
      (#f other-type-bits other-type-bits)])

   (make-prim (char->integer)
     [(fxnum-type-bits bot-type)
      (#f char-type-bits char-type-bits)])

   (make-prim (char-alphabetic? char-numeric? char-whitespace? char-upper-case?
               char-lower-case? char-title-case?)
     [(true-type-bits false-type-bits)
      (#f char-type-bits char-type-bits)])

   (make-prim (char-general-category)
     [(symbol-type-bits bot-type)
      (#f char-type-bits char-type-bits)])

   (make-prim (integer->char)
     [(char-type-bits bot-type)
      (#f fxnum-type-bits fxnum-type-bits)])

   (make-prim (char=? char<=? char<? char>? char>=?)
     [(true-type-bits false-type-bits)
      (#f char-type-bits char-type-bits)
      (#f char-type-bits char-type-bits) ...])

   (make-prim (char-upcase char-downcase char-titlecase char-foldcase)
     [(char-type-bits bot-type)
      (#f char-type-bits char-type-bits)])

   (make-prim (string?)
     [(true-type-bits false-type-bits)
      (#f string-type-bits (not-type string-type-bits))])

   (make-prim (open-string-input-port)
     [(other-type-bits bot-type)
      (#f string-type-bits string-type-bits)])

   (make-prim (list->string)
     [(string-type-bits bot-type)
      (#f pair-or-null-type-bits pair-or-null-type-bits)])

   (make-prim (string->list)
     [(pair-or-null-type-bits bot-type)
      (#f string-type-bits string-type-bits)])

   (make-prim (string-append)
     [(string-type-bits bot-type)
      (#f string-type-bits string-type-bits) ...])

   (make-prim (string-length)
     [(fxnum-type-bits bot-type)
      (#f string-type-bits string-type-bits)])

   (make-prim (string-ref)
     [(char-type-bits bot-type)
      (#f string-type-bits string-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)])

   (make-prim (string-set!)
     [(true-type bot-type)
      (#f string-type-bits string-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f char-type-bits char-type-bits)])

   (make-prim (string)
     [(string-type-bits bot-type)
      (#f char-type-bits char-type-bits) ...])

   (make-prim (r6rs:char=? r6rs:char<=? r6rs:char<? r6rs:char>? r6rs:char>=?)
     [(true-type-bits false-type-bits)
      (#f char-type-bits char-type-bits)
      (#f char-type-bits char-type-bits)
      (#f char-type-bits char-type-bits) ...])

   (make-prim (r6rs:string=? r6rs:string<? r6rs:string<=? r6rs:string>=? r6rs:string>?)
     [(true-type-bits false-type-bits)
      (#f string-type-bits string-type-bits)
      (#f string-type-bits string-type-bits)
      (#f string-type-bits string-type-bits) ...])

   (make-prim (string=? string<? string<=? string>=? string>?)
     [(true-type-bits false-type-bits)
      (#f string-type-bits string-type-bits)
      (#f string-type-bits string-type-bits) ...])

   (make-prim (string->symbol)
     [(symbol-type-bits bot-type)
      (#f string-type-bits string-type-bits)])

   (make-prim (substring)
     [(string-type-bits bot-type)
      (#f string-type-bits string-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)])

   (make-prim (string-normalize-nfd string-normalize-nfkd
               string-normalize-nfc string-normalize-nfkc)
     [(string-type-bits bot-type)
      (#f string-type-bits string-type-bits)])

   (make-prim (make-string)
     [(string-type-bits bot-type)
      (#f fxnum-type-bits fxnum-type-bits)]
     [(string-type-bits bot-type)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f char-type-bits char-type-bits)])

   ;; Can't use unknown-prim b/c that would escape it's arguments
   ;; for Chez we my be more precise and say void instead of top-type
   (make-prim (set-car! set-cdr!) [(top-type top-type)
                                   (#f pair-type-bits pair-type-bits)
                                   (#t top-type top-type)])

   (make-prim (length) [(length-type-bits bot-type)
                        (#f pair-or-null-type-bits pair-or-null-type-bits)])

   ;; vector related primitives
   (make-pred-prim (vector?) vector-type-bits)

   (make-top-prim (vector) [(vector-type-bits bot-type) #t ...])

   (make-prim (vector->list)
     [(pair-or-null-type-bits bot-type)
      (#f vector-type-bits vector-type-bits)])

   (make-prim (make-vector)
     [(vector-type-bits bot-type) (#f length-type-bits length-type-bits)]
     [(vector-type-bits bot-type) (#f length-type-bits length-type-bits)
                                  (#t top-type top-type)])

   (make-prim (vector-length) [(length-type-bits bot-type)
                               (#f vector-type-bits vector-type-bits)])

   (make-prim (vector-ref) [(true-type false-type) 
                            (#f vector-type-bits vector-type-bits)
                            (#f length-type-bits length-type-bits)])

   ;; for Chez we may be more precise and say void instead of top-type
   (make-prim (vector-set!) [(true-type bot-type)
                             (#f vector-type-bits vector-type-bits)
                             (#f length-type-bits length-type-bits)
                             (#t top-type top-type)])

   ;; general integer related primitives
   ;;(make-pred-prim (integer?) integer-type-bits)

   (make-pred-prim (bignum?) bignum-type-bits)

   ;; +0.0, +0.0i, etc. are also zero so we have to include other-type-bits
   (make-pred-prim (zero?) (bitwise-ior fxzero-type-bits other-number-type-bits))
   (make-prim (fxzero?)
     [(true-type-bits false-type-bits)
      (#f fxzero-type-bits (bitwise-xor fxnum-type-bits fxzero-type-bits))])

   ;; fixnum related primitives
   ;; TODO: if/when we "cheat" and make overflow checks explicit, these will
   ;;       need to have a return type of integer-type-bits and we'll rely on the
   ;;       check afterwards cleaning up for us.
   (make-pred-prim (fixnum?) fxnum-type-bits)

   (make-prim (fx+ fx* fxxor)
     [(fxnum-type-bits bot-type) 
      (#f fxnum-type-bits fxnum-type-bits) ...])

   (make-prim (fx- fx/)
     [(fxnum-type-bits bot-type) 
      (#f fxnum-type-bits fxnum-type-bits)]
     [(fxnum-type-bits bot-type) 
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits) ...])

   (make-prim (fxarithmetic-shift-left fxarithmetic-shift-right fxarithmetic-shift)
     [(fxnum-type-bits bot-type)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)]) ; TODO: add a way to indicate sub-ufixnum

   ;;; TODO: defines prim boxes for usafe-fx<=, fx<, fx=, fx>, fx>= that will
   ;;;       better embody the more specific information we're learning
   ;;;       from them.
   ; fx<=
   ; fx<
   ; fx=
   ; fx>=
   ; fx>
   ;; TODO: why does Chez have both 'fx<' and 'fx<?'
   ;; TODO: why does Chez allow only one argument but not zero arguments?
   (make-prim (fx<= fx< fx= fx>= fx>)
     [(true-type-bits false-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits) ...])

   (make-prim (fx<? fx=? fx<=? fx>=? fx>?)
     [(true-type-bits false-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits)
      (#f fxnum-type-bits fxnum-type-bits) ...])

   (make-pred-prim (flonum?) flonum-type-bits)

   (make-prim (fl+ fl*) [(flonum-type-bits bot-type)
                         (#f flonum-type-bits flonum-type-bits) ...])

   (make-prim (fldiv flmod fldiv0 flmod0 flexpt)
     [(flonum-type-bits bot-type)
      (#f flonum-type-bits flonum-type-bits)
      (#f flonum-type-bits flonum-type-bits)])

   (make-prim (flatan fllog)
     [(flonum-type-bits bot-type)
      (#f flonum-type-bits flonum-type-bits)]
     [(flonum-type-bits bot-type)
      (#f flonum-type-bits flonum-type-bits)
      (#f flonum-type-bits flonum-type-bits)])

   (make-prim (flsqrt flabs flnumerator fldenominator flfloor flceiling
               fltruncate flround flexp flsin flcos fltan flasin flacos)
     [(flonum-type-bits bot-type)
      (#f flonum-type-bits flonum-type-bits)])

   (make-prim (fl- fl/ flmax flmin)
     [(flonum-type-bits bot-type)
      (#f flonum-type-bits flonum-type-bits)]
     [(flonum-type-bits bot-type)
      (#f flonum-type-bits flonum-type-bits)
      (#f flonum-type-bits flonum-type-bits) ...])

   (make-prim (fl>? fl<? fl>=? fl<=? fl=?)
     [(true-type-bits false-type-bits)
      (#f flonum-type-bits flonum-type-bits)
      (#f flonum-type-bits flonum-type-bits)
      (#f flonum-type-bits flonum-type-bits) ...])

   (make-prim (flpositive? flnegative? flzero? flinteger? flodd? fleven?
               flfinite? flinfinite? flnan?)
     [(true-type-bits false-type-bits)
      (#f flonum-type-bits flonum-type-bits)])

   (make-pred-prim (number?) number-type-bits)

   ;; other type-predicate primitives
   (make-pred-prim (boolean?) boolean-type-bits)

   (make-pred-prim (symbol?) symbol-type-bits)

   (make-pred-prim (null?) null-type-bits)

   ;; TODO: can't use make-pred-prim b/c that is only for non-proc-type-bits
   (make-prim (procedure?)
     [(true-type-bits false-type-bits)
      (#f proc-type non-proc-type-bits)])
   
   ;;; TODO: defines prim box for eq? that will embody the more specific
   ;;; information we're learning from eq? 
   ;;; TODO: Add to case-lambda the ability to flag it as having been used
   ;;;       in an eq? so we need to preserve it, even if we could normally
   ;;;       reduce it.
              
   ;; TODO: when we add ports, we'll need a display, and need to make
   ;;       the checks explicit for display
   (make-top-prim (display)
     ;; for Chez we may be more precise and say void instead of top-type
     [(true-type false-type) #f]
     [(true-type false-type) #f #f])

   ;; if void is really unspecified, it should be top here.
   (make-top-prim (void) [(other-type-bits bot-type)])

   ;; TODO: when we add ports, we'll need an read, and need to make
   ;;       the checks explicit for read
   (make-top-prim (read)
     ;; TODO: do we need to type-intersect the return values?
     [(non-proc-type-bits non-proc-type-bits)]
     [(non-proc-type-bits non-proc-type-bits) #f])

   ;; The make-bottom that always returns a bottom type
   (make-top-prim (make-bottom) [(bot-type bot-type)])

   (list (cons 'unknown-prim (lambda (ast) top-type))) ;; prims could be non-procs

   ))

(define (prim-name->type ast name)
  (let ([prim (assq name prim-table)])
    (if prim
        ((cdr prim) ast)
        (prim-name->type ast 'unknown-prim))))

)
