(library (tests)
  (export run-tests)
  (import (rnrs) (match) (driver-function)
          (only (chezscheme) display-condition pretty-print))

(define (alpha-equiv? x y env)
  (define (alpha-equiv-var? x y)
    (let ([ypair (assq y env)])
      (if ypair (eq? x (cdr ypair)) (eq? x y))))
  (define (alpha-equiv-clause clx cly)
    (match (list clx cly)
      [(((,xx* ...) ,xb) ((,yx* ...) ,yb))
       (let ([env (append (map cons yx* xx*) env)])
         (alpha-equiv? xb yb env))]
      [(((,xx* ... . ,xr) ,xb) ((,yx* ... . ,yr) ,yb))
       (let ([env (cons (cons yr xr) (append (map cons yx* xx*) env))])
         (alpha-equiv? xb yb env))]))
  (match `(,x ,y)
    [((lambda . ,clx) (lambda . ,cly)) (alpha-equiv-clause clx cly)]
    [((case-lambda ,clx* ...) (case-lambda ,cly* ...))
     (for-all alpha-equiv-clause clx* cly*)]
    [((letrec ([,xx* ,xe*] ...) ,xb) (letrec ([,yx* ,ye*] ...) ,yb))
     (let ([env (append (map cons yx* xx*) env)])
       (and (for-all (lambda (x y) (alpha-equiv? x y env)) xe* ye*)
            (alpha-equiv? xb yb env)))]
    [((set! ,xx ,xv) (set! ,yx ,yv))
     (and (alpha-equiv-var? xx yx) (alpha-equiv? xv yv env))]
    [(,x ,y) (guard (and (symbol? x) (symbol? y))) (alpha-equiv-var? x y)]
    [((begin ,xa ,xb) (begin ,ya ,yb))
     (and (alpha-equiv? xa ya env) (alpha-equiv? xb yb env))]
    [((quote ,x) (quote ,y)) (equal? x y)]
    [((if ,xt ,xc ,xa) (if ,yt ,yc ,ya))
     (and (alpha-equiv? xt yt env)
          (alpha-equiv? xc yc env)
          (alpha-equiv? xa ya env))]
    [((prim-wrap ,xchanged? ,xprim ,xb) (prim-wrap ,ychanged? ,yprim ,yb))
     (and (eq? xprim yprim) (alpha-equiv? xb yb env) (eq? xchanged? ychanged?))]
    [((,xrator ,xrand* ...) (,yrator ,yrand* ...))
     (and (alpha-equiv? xrator yrator env)
          (= (length xrand*) (length yrand*))
          (for-all (lambda (x y) (alpha-equiv? x y env)) xrand* yrand*))]
    [(,x ,y) #f]))

(define tests '(
  [1 . '1]
  [(cons 1 2) . (cons '1 '2)]
  [(car (cons 1 2)) . (unsafe-car (cons '1 '2))]
  [(lambda () 1) . (lambda () '1)]
  [(lambda (x) x) . (lambda (x) x)]
  [(lambda x x) . (lambda x x)]
  [(lambda (x . y) x) . (lambda (x . y) x)]
  [(lambda (x . y) y) . (lambda (x . y) y)]
  [(let ([z (lambda (x) x)]) (let ([y (z (cons 1 2))]) (car y))) .
   (letrec ([z (lambda (x) x)]) (letrec ([y (z (cons '1 '2))]) (unsafe-car y)))]
  [(let ([z (lambda (f y) (f y))])
     (let ([y (z (lambda (x) x) (cons 1 2))]) (car y))) .
   (letrec ([z (lambda (f y) (f y))])
     (letrec ([y (z (lambda (x) x) (cons '1 '2))]) (unsafe-car y)))]
  [(cons #f (case-lambda
             [(x y z) (cons x (cons y z))]
             [(x . r) (cons r x)]
             [q (car q) (cdr q)])) .
   (cons '#f (case-lambda
              [(x y z) (cons x (cons y z))]
              [(x . r) (cons r x)]
              [q (car q)]))]
  [(let ([x (cons 1 2)]) (car x)) . (letrec ([x (cons '1 '2)]) (unsafe-car x))]
  [(let ([x (cons 1 2)]) (car x) (car x)) .
   (letrec ([x (cons '1 '2)]) (begin (unsafe-car x) (unsafe-car x)))]
  [(let ([x (cons 1 2)]) (car x) (vector-length x)) .
   (letrec ([x (cons '1 '2)]) (begin (unsafe-car x) (vector-length x)))]
  [(let ([x (read)]) (car x) (car x)) .
   (letrec ([x (read)]) (begin (car x) (unsafe-car x)))]
  [(if (error 'who "message") 1 2) . (error 'who '"message")]
  [(if #t 1 2) . (begin '#t '1)] ;; test unreachable empty environments
  [(let ((f (lambda (x) (if (fxzero? x) 1 2)))) (f 0) (f 3)) . 
   (letrec ([f (lambda (x) (if (fxzero? x) '1 '2))]) (begin (f '0) (f '3)))]
  ;;; Olin Shiver's puzzle test (page 140 (in pdf - 118) of his thesis):
  [(let ([f (lambda (x h) (if (fxzero? x) (h) (lambda () x)))]) (f 0 (f 3 #f))) .
   (letrec ([f (lambda (x h) (if (fxzero? x) (h) (lambda () x)))])
     (f '0 (f '3 '#f)))]
  [(letrec ([x (cons 1 2)] [f (lambda () (car x))]) (f)) .
   (letrec ([x (cons '1 '2)] [f (lambda () (unsafe-car x))]) (f))]
  ;; Test that bare var refs properly effect the environment
  [(let ([x (if (read) (cons 1 2) '#f)]) (if x (car x) 3)) .
   (letrec ([x (if (read) (cons '1 '2) '#f)]) (if x (unsafe-car x) '3))]
  [(let ([x (read)]) (let ([a (car x)]) (car x))) .
   (letrec ([x (read)]) (letrec ([a (car x)]) (unsafe-car x)))]
  [(let ([x (read)]) (if (if (pair? x) #f #t) (car x) (car x))) .
   (letrec ([x (read)]) (if (if (pair? x) '#f '#t) (car x) (unsafe-car x)))]
  [(letrec ([a (car (cons 1 2))]) 1) . (letrec ([a (unsafe-car (cons '1 '2))]) '1)]
  [(letrec ((x (cons 1 2))) (car x)) . (letrec ([x (cons '1 '2)]) (unsafe-car x))]
  [(let ([x (cons 1 2)]) (letrec ([a (car x)][b (car x)]) 1)) .
   (letrec ([x (cons '1 '2)]) (letrec ([a (unsafe-car x)][b (unsafe-car x)]) '1))]
  ;; the cons call to f never happens
  [(let ([f (lambda (x) (car x) 1)]) (f (make-vector 3)) (f (cons 1 2))) .
   (letrec ([f (lambda (x) (car x))]) (f (unsafe-make-vector '3)))]
  ;; same as last test but both calls are made because of different order
  [(let ([f (lambda (x) (car x) 1)]) (f (cons 1 2)) (f (make-vector 3)) 4) .
   (letrec ([f (lambda (x) (begin (car x) '1))])
     (begin (f (cons '1 '2)) (f (unsafe-make-vector '3))))]
  [(car (read)) . (car (read))]
  [(cons 1 2) . (cons '1 '2)]
  [(read) . (read)]
  [(letrec ([a (car b)][b (car a)]) 3) .
   (letrec ([a (unsafe-car b)][b (unsafe-car a)]) '3)]
  [(letrec ([a (lambda () (car b))][b (lambda () (car a))]) 3) .
   (letrec ([a (case-lambda)] [b (case-lambda)]) '3)]
  [(letrec ([a (cons 1 2)] [b (car a)]) 3) .
   (letrec ([a (cons '1 '2)][b (unsafe-car a)]) '3)]
  [(let ([a (lambda () (car (read)))]) 1) .
   (letrec ([a (case-lambda)]) '1)]
  [(begin (error 'who "message") #t) . (error 'who '"message")]
  ;; tests that info is gathered from both args of letrec
  [(let ([x (read)]) (letrec ([a (car x)][b (vector-length x)]) 4 5)) .
   (letrec ([x (read)]) (letrec ([a (car x)][b (vector-length x)]) '4))]
  [(let ([x (lambda () 1 2)]) 3) . (letrec ([x (case-lambda)]) '3)]
  [(let ([x (read)]) (if x x (error 'who "message")) (if x 1 2)) .
   (letrec ([x (read)]) (begin (if x x (error 'who '"message")) (begin x '1)))]
  [(let ([x (cons 1 2)]) (car x) (set! x 5) (car x)) .
   (letrec ([x (cons '1 '2)]) (begin (begin (car x) (set! x '5)) (car x)))]
  [(let ([x (cons 1 2)]) (car x) (set! x (cons 3 4)) (car x)) .
   (letrec ([x (cons '1 '2)])
     (begin (begin (unsafe-car x) (set! x (cons '3 '4))) (unsafe-car x)))]
  [(let ([f (lambda (x) (set! x (cons 1 2)) x)]) (car (f (cons 4 5)))) .
   (letrec ([f (lambda (x) (begin (set! x (cons '1 '2)) x))])
     (unsafe-car (f (cons '4 '5))))]
  ;; The following works, should reach the 5, since (cons - -) is true
  [(let ([f (lambda (x y) (cons (car x) (fx+ y 1)))] [x (read)] [y (read)])
     (if (f x y) 5 7)) .
   (letrec ([f (lambda (x y) (cons (car x) (fx+ y '1)))] [x (read)] [y (read)])
     (begin (f x y) '5))]
  [(let ([x (read)] [f (lambda (a b) (cdr a))]) (car x) (f x (car x))) .
   (letrec ([x (read)] [f (lambda (a b) (unsafe-cdr a))])
     (begin (car x) (f x (unsafe-car x))))]
  [(let ([x (read)][f (lambda (a b) (cdr a))]) (f (begin (car x) x) (car x))) .
   (letrec ([x (read)] [f (lambda (a b) (unsafe-cdr a))])
     (f (begin (car x) x) (car x)))]
  [(let ([x (read)]) (if #t (car x) 1) (car x)) .
   (letrec ([x (read)]) (begin (begin '#t (car x)) (unsafe-car x)))]
  [(letrec ([x 1]) (set! x (cons 1 2)) (car x)) .
   (letrec ([x '1]) (begin (set! x (cons '1 '2)) (car x)))]
  ;; check that escaping lambdas work right
  [(if (read) (lambda (x) (car x)) (lambda (y) (cdr y))) .
   (if (read) (lambda (x) (car x)) (lambda (y) (cdr y)))]
  ;; should mark-escaped-lambda
  [(cons 1 (lambda (x) (cdr x))) . (cons '1 (lambda (x) (cdr x)))]
  ;; should not mark-escaped-lambda
  [(null? (lambda (x) (cdr x))) . (null? (case-lambda))]
  ;; test that car never happens due to type error in cons call
  [(car (cons 1)) . (unsafe-car (cons '1))]
  ;; test that low gas works ok
  [(let ([x (read)]) (if (not (not (pair? x))) (car x))) .
   (letrec ([x (read)])
     (if (if (if (pair? x) '#f '#t) '#f '#t) (unsafe-car x) (void)))]
  [(lambda () (void) 3) . (lambda () (begin (void) '3))]
  [(letrec ([fac (lambda (x) (if (fxzero? x) 1 (fx* x (fac 0))))]) (fac (read))) .
   (letrec ([fac (lambda (x) (if (fxzero? x) '1 (unsafe-fx* x (fac '0))))])
     (fac (read)))]
  ;; test incorrect argument count calls
  [((lambda (x) 1 2) 3 4) . ((case-lambda) '3 '4)]
  ;; too many arguments and we have a last rest (should call body)
  [((case-lambda [(x . y) 3 4]) 5 6 7) . ((lambda (x . y) (begin '3 '4)) '5 '6 '7)]
  ;; too many argument and no last-rest
  [((case-lambda [(x y) 3 4]) 5 6 7) . ((case-lambda) '5 '6 '7)]
  ;; small number of arguments and exact call but no clause handles it
  [((case-lambda [(x y z w) 3 4]) 5 6 7) . ((case-lambda) '5 '6 '7)]
  ;; small number of argument and exact call and a clause handles it
  [((case-lambda [(x y z w) 3 4]) 5 6 7 8) .
   (letrec ([x '5][y '6][z '7][w '8]) (begin '3 '4))]
  [((lambda (x . y) 3) 1 2) . ((lambda (x . y) '3) '1 '2)]
  ))

(define (run-tests)
  (let ([len (length tests)])
    (let f ([tests tests] [failed '()])
      (cond
       [(null? tests)
        (if (null? failed)
            (begin
              (display "\nran ")
              (display len)
              (display " tests, all succeeded\n"))
            (begin
              (display "\nran ")
              (display len)
              (display " tests, ")
              (display (length failed))
              (display " failed\n")
              (display "Failing tests:\n")
              (for-each pretty-print (reverse failed))))]
       [else
        (let ([fail (lambda ()
                      (display "\nfailed: ")
                      (display (caar tests))
                      (newline)
                      (f (cdr tests) (cons (caar tests) failed)))])
          (guard
           (condition [#t (newline) (display-condition condition) (fail)])
           (let ([result (run (caar tests))])
             (cond
              [(alpha-equiv? (cdar tests) result '())
               (display ".")
               (f (cdr tests) failed)]
              [else (fail)]))))]))))

)
