(library (types)
  (export
   select-CFA
   type? ;; obj -> bool
   type=? ;; type * type -> bool
   type<=? ;; type * type -> bool (true if the type is = or at a lower point on the lattice)
  proc-type-bits=? ;; list-of type * list-of type -> bool (true if the first and second set of procs contain the same information)

  ;; non-proc type bits (fixnum)
  non-proc-type-bits
  other-type-bits
  false-type-bits
  true-type-bits
  pair-or-null-type-bits
  null-type-bits
  pair-type-bits
  symbol-type-bits
  boolean-type-bits
  number-type-bits
  fxnum-type-bits
  flonum-type-bits
  other-number-type-bits
  exact-integer-type-bits
  fxzero-type-bits
  bignum-type-bits
  length-type-bits
  vector-type-bits
  string-type-bits
  char-type-bits

  ;; proc-type-bits
  type-procs ;; type -> proc-type-bits
  make-proc-type
  make-proc-record
  proc-type-bits-case
  bot-proc-type-bits ;; void -> proc-type-bits
  top-proc-type-bits ;; void -> proc-type-bits
  one-or-more-proc-type-bits ;; proc-type -> proc-type-bits

  ;; proc-record
  proc-record->type ;; proc-record -> type
  proc-record-input-called-proc ;; proc-record -> ???
  proc-record-input-argument-proc ;; proc-record -> ???
  proc-record-output-return-proc ;; proc-record -> ???
  proc-record-output-argument-proc ;; proc-record -> ???

  ;; top/bot related types
  top-type ;; type
  bot-type ;; type
  bot-type? ;; type -> bool
  non-bot-type? ;; type -> bool

  ;; type predicates and constants
  proc-type ;; type
  true-type true-type? ;; type, type->bool
  false-type false-type? ;; type, type->bool
  pair-type? null-type? ;; type, type->bool

  ;; type operators (type * type -> type)
  type-union
  type-intersect

  maybe-non-proc-type->type
  not-type
  type-of
  mark-escaped-type!
  type-filter
  )
 (import (except (rnrs) case cond if) (safe-forms) (records-io) (datatype)
         (only (chezscheme) void bignum?))

;; Choose to use 0CFA or Sub-0CFA at compile-time and avoid run-time dispatch
(define-syntax select-CFA
  (syntax-rules (zero-CFA sub-zero-CFA)
    [(_ [(zero-CFA) zcfa zcfa* ...] [(sub-zero-CFA) sub-zcfa sub-zcfa* ...])
     (begin sub-zcfa sub-zcfa* ...)]   ;; change to sub-zcfa to choose sub-zero CFA
    [(_ [(sub-zero-CFA) sub-zcfa sub-zcfa* ...] [(zero-CFA) zcfa zcfa* ...])
     (begin sub-zcfa sub-zcfa* ...)])) ;; change to sub-zcfa to choose sub-zero CFA

;; Full types
(define-record-type type (fields non-proc procs)
  (nongenerative)
  (protocol
   (lambda (make)
     (lambda (non-proc procs)
       (assert (fixnum? non-proc))
       (assert (proc-type-bits? procs))
       (make non-proc procs)))))

(define (type=? x y)
  (and (= (type-non-proc x) (type-non-proc y))
       (proc-type-bits=? (type-procs x) (type-procs y))))

(define (type<=? x y)
  (or (type=? x y) (type=? x (type-intersect x y))))

;; Non-proc types
(define-syntax flatten-mask-bits ;; (a b (c d e)) -> (bitwise-ior a b c d e)
  (syntax-rules ()
    [(_ (arg sub-args ...) args ...)
     (bitwise-ior arg (flatten-mask-bits args ...))]
    [(_ arg args ...)
     (bitwise-ior arg (flatten-mask-bits args ...))]
    [(_) 0]))

(define-syntax define-mask-bits
  (syntax-rules ()
    [(_ offset (arg sub-args ...) args ...)
     (begin
       (define-mask-bits offset sub-args ... args ...)
       (define arg (flatten-mask-bits sub-args ...)))]
    [(_ offset arg) (define arg (bitwise-arithmetic-shift 1 offset))]
    [(_ offset arg args ...)
     (begin
       (define arg (bitwise-arithmetic-shift 1 offset))
       (define-mask-bits (+ 1 offset) args ...))]))

(define-mask-bits 0
  (non-proc-type-bits

   ;; Singleton types
   (false-non-proc-type-bits
    false-type-bits)

   (true-non-proc-type-bits
    other-type-bits ;; handles types not listed here (e.g. unspecified type)
    true-type-bits

    ;; Container types
    (pair-or-null-type-bits
     null-type-bits ;; not exactly a container type, but pair-or-null-type is.
     pair-type-bits)
    vector-type-bits

    ;; Atomic types
    string-type-bits
    char-type-bits
    symbol-type-bits ;; TODO: property lists
    (number-type-bits
     (other-number-type-bits
      flonum-type-bits)
     (exact-integer-type-bits
      (bignum-type-bits
       neg-bignum-type-bits
       pos-bignum-type-bits)
      (fxnum-type-bits
       min-fxnum-type-bits
       neg-index-type-bits
       neg-one-type-bits
       (length-type-bits
        fxzero-type-bits
        one-type-bits
        index-type-bits
        max-fxnum-type-bits)))))))

(define boolean-type-bits (bitwise-ior true-type-bits false-type-bits))

;; Proc types

;; For type that hasn't yet been discovered to be proc.
(define-syntax bot-proc-type-bits
  (select-CFA
   [(zero-CFA) (syntax-rules () [(_) '()])]
   [(sub-zero-CFA) (syntax-rules () [(_) #f])]))

;; For procedure that has been discovered to be the union of two
;; different procs.
(define-syntax top-proc-type-bits
  (select-CFA
   [(zero-CFA) (syntax-rules () [(_) (list #t)])]
   [(sub-zero-CFA) (syntax-rules () [(_) #t])]))

;; Other proc types point directly to a lambda box
(define-syntax one-or-more-proc-type-bits
  (select-CFA
   [(zero-CFA) (syntax-rules () [(_ x x* ...) (list x x* ...)])]
   [(sub-zero-CFA) (syntax-rules () [(_ x) x])]))

(define (proc-type-bits? x)
  (select-CFA
   [(zero-CFA) (for-all (lambda (x) (or (proc-record? x) (eq? #t x))) x)]
   [(sub-zero-CFA) (or (eq? x #t) (eq? x #f) (proc-record? x))]))

(define-syntax proc-type-bits-case
  (syntax-rules (bot-proc-type-bits top-proc-type-bits one-or-more-proc-type-bits)
    [(_ proc-type-bits
        [(bot-proc-type-bits) bot ...]
        [(top-proc-type-bits) top ...]
        [(one-or-more-proc-type-bits procs) one-or-more ...])
     (select-CFA
      [(zero-CFA)
       (let ([procs proc-type-bits])
         (cond
          [(eq? procs '()) bot ...]
          [(memq #t procs) top ...]
          [else one-or-more ...]))]
      [(sub-zero-CFA)
       (let ([procs proc-type-bits])
         (cond
          [(eq? procs #f) bot ...]
          [(eq? procs #t) top ...]
          [else one-or-more ...]))])]))

(define (proc-type-bits=? x y)
  (select-CFA
   [(zero-CFA)
    ;; ugly -- doing the quadratic check to see if two sets of procs are the same.
    (let f ([x* x] [y* y] [ry* '()] [looped? #f])
      (cond
       [(and (null? x*) (null? y*) (null? ry*)) #t]
       [(or (null? x*) (and (null? y*) (null? ry*)) (and (null? y*) looped?)) #f]
       [(null? y*) (f x* ry* '() #t)]
       [(eq? (car x*) (car y*)) (f (cdr x*) (cdr y*) ry* #f)]
       [else (f x* (cdr y*) (cons (car y*) ry*) looped?)]))]
   [(sub-zero-CFA) (eq? x y)]))

(define-record-type proc-record
  (nongenerative)
  (fields
   escape-proc ;; void -> void
   input-called-proc ;; arg-count * bool -> void
   input-argument-proc ;; arg-count * position * type -> void
   output-return-proc ;; arg-count -> flow-graph-node[type]
   output-argument-proc ;; arg-count * position -> flow-graph-node[bitype]
   ))

(define (make-proc-type escape
                        input-called input-argument
                        output-return output-argument)
  (make-type 0 (one-or-more-proc-type-bits
                (make-proc-record
                 escape input-called input-argument
                 output-return output-argument))))

(define (proc-record->type x) (make-type 0 x))

;; Type constants and predicates
(define top-type (make-type non-proc-type-bits (top-proc-type-bits)))
(define bot-type (make-type 0 (bot-proc-type-bits)))
(define (bot-type? x) (assert (type? x)) (type=? bot-type x))
(define (non-bot-type? x) (assert (type? x)) (not (bot-type? x)))

(define proc-type (make-type 0 (top-proc-type-bits)))
(define true-type (make-type true-non-proc-type-bits (top-proc-type-bits)))
(define (true-type? x) (non-bot-type? (type-intersect x true-type)))
(define false-type (make-type false-non-proc-type-bits (bot-proc-type-bits)))
(define (false-type? x) (non-bot-type? (type-intersect x false-type)))

(define pair-type (make-type pair-type-bits (bot-proc-type-bits)))
(define (pair-type? x) (non-bot-type? (type-intersect x pair-type)))
(define null-type (make-type null-type-bits (bot-proc-type-bits)))
(define (null-type? x) (non-bot-type? (type-intersect x null-type)))

(define (type-intersect x y)
  (make-type
   (bitwise-and (type-non-proc x) (type-non-proc y))
   (select-CFA
    [(zero-CFA) (proc-type-bits-intersect (type-procs x) (type-procs y))]
    [(sub-zero-CFA)
     (proc-type-bits-case (type-procs x)
       [(bot-proc-type-bits) (bot-proc-type-bits)]
       [(top-proc-type-bits) (type-procs y)]
       [(one-or-more-proc-type-bits proc-x)
        (proc-type-bits-case (type-procs y)
          [(bot-proc-type-bits) (bot-proc-type-bits)]
          [(top-proc-type-bits) (type-procs x)]
          [(one-or-more-proc-type-bits proc-y)
           (if (eq? proc-x proc-y) (type-procs x) (bot-proc-type-bits))])])])))

(define (proc-type-bits-intersect x* y*)
  (cond
   [(memq #t x*) y*]
   [(memq #t y*) x*]
   [else (let f ([x* x*] [final* '()])
           (if (null? x*)
               final*
               (let ([x (car x*)])
                 (if (memq x y*)
                     (f (cdr x*) (cons x final*))
                     (f (cdr x*) final*)))))]))

(define (proc-type-bits-union x* y*)
  (let f ([x* x*] [xy* y*])
    (if (null? x*)
        xy*
        (let ([x (car x*)])
          (if (memq x y*)
              (f (cdr x*) xy*)
              (f (cdr x*) (cons x xy*)))))))

;; NOTE: may side-effect escape value of procs
(define (type-union x y)
  (select-CFA
   [(zero-CFA)
    (define (mark-escaped-lambda! proc*)
      (map (lambda (proc) (unless (eq? proc #t) ((proc-record-escape-proc proc))))
           proc*))
    (let ([proc-x* (type-procs x)] [proc-y* (type-procs y)])
      (when (memq #t proc-x*) (mark-escaped-lambda! proc-y*))
      (when (memq #t proc-y*) (mark-escaped-lambda! proc-x*)))
    (make-type
     (bitwise-ior (type-non-proc x) (type-non-proc y))
     (proc-type-bits-union (type-procs x) (type-procs y)))]
   [(sub-zero-CFA)
    (define (mark-escaped-lambda! proc) ((proc-record-escape-proc proc)))
    (make-type
     (bitwise-ior (type-non-proc x) (type-non-proc y))
     (proc-type-bits-case (type-procs x)
       [(bot-proc-type-bits) (type-procs y)]
       [(top-proc-type-bits)
        (proc-type-bits-case (type-procs y)
          [(bot-proc-type-bits) (values)]
          [(top-proc-type-bits) (values)]
          [(one-or-more-proc-type-bits proc-y) (mark-escaped-lambda! proc-y)])
        (top-proc-type-bits)]
       [(one-or-more-proc-type-bits proc-x)
        (proc-type-bits-case (type-procs y)
          [(bot-proc-type-bits) (type-procs x)]
          [(top-proc-type-bits) (mark-escaped-lambda! proc-x) (top-proc-type-bits)]
          [(one-or-more-proc-type-bits proc-y)
           (if (eq? proc-x proc-y)
               (type-procs x)
               (begin
                 (mark-escaped-lambda! proc-x) ;; rather then escaping keep both
                 (mark-escaped-lambda! proc-y)
                 (top-proc-type-bits)))])]))]))

(define (mark-escaped-type! type)
  (select-CFA
   [(zero-CFA)
    (for-each 
     (lambda (proc) (when (proc-record? proc) ((proc-record-escape-proc proc))))
     (type-procs type))]
   [(sub-zero-CFA) (type-union type top-type)])
  (values))

(define (type-filter type true false)
  (type-union
   (if true (type-intersect type true-type) bot-type)
   (if false (type-intersect type false-type) bot-type)))

;; helpers for prims
;; TODO: rename as type-complement
(define (not-type non-proc-type-bits)
  (make-type (bitwise-not non-proc-type-bits) (top-proc-type-bits)))
;; TODO: make work for proc-type
(define (maybe-non-proc-type->type type)
  (if (type? type) type (make-type type (bot-proc-type-bits))))

(define (type-of x)
  (make-type
   (cond
    [(eq? x (void)) other-type-bits]
    [(eq? x #f) false-type-bits]
    [(eq? x #t) true-type-bits]
    [(eq? x '()) null-type-bits]
    [(pair? x) pair-type-bits]
    [(vector? x) vector-type-bits]
    [(symbol? x) symbol-type-bits]
    [(eq? x 0) fxzero-type-bits]
    [(eq? x 1) one-type-bits]
    [(eq? x -1) neg-one-type-bits]
    [(and (not (bignum? x)) (fixnum? x) (bignum? (+ x 1))) max-fxnum-type-bits]
    [(and (not (bignum? x)) (fixnum? x) (bignum? (- x 1))) min-fxnum-type-bits]
    [(and (fixnum? x) (positive? x)) index-type-bits]
    [(and (fixnum? x) (negative? x)) neg-index-type-bits]
    [(and (bignum? x) (positive? x)) pos-bignum-type-bits]
    [(and (bignum? x) (negative? x)) neg-bignum-type-bits]
    [(string? x) string-type-bits]
    [(procedure? x) 0]
    [(flonum? x) flonum-type-bits]
    [(char? x) char-type-bits]
    [else non-proc-type-bits])
   (if (procedure? x) (top-proc-type-bits) (bot-proc-type-bits))))

(record-writer (record-type-descriptor type)
  (lambda (r p wr)
    (display "#<0x" p)
    (display (number->string (type-non-proc r) 16) p)
    (display ", " p)
    (wr (proc-type-bits-case (type-procs r)
          [(bot-proc-type-bits) 'bot]
          [(top-proc-type-bits) 'top]
          [(one-or-more-proc-type-bits proc) 'one-or-more-proc]) p)
    (display ">" p)))

(record-writer (record-type-descriptor proc-record)
  (lambda (r p wr) (display "#[proc-type]" p)))

)
