(library (value-flow-graphs)
  (export make-value-flow-graph clause-info-in-called)
  (import (except (rnrs) case cond if) (safe-forms) (define-ignored) (ast)
          (ast-flow-graphs) (flow-graphs) (types) (bitypes) (idents)
          (composition-stacks) (env-flow-functions) (prims) (iota) (ast-bubbles))

(define make-AST-for-all-node
  (case-lambda
   [(ast name fun base nodes1 nodes2)
    (cond
     [(null? nodes1) base]
     [else
      (make-AST-for-all-node
       ast name fun (make-AST-node ast make-flow-graph-function-node name boolean=?
                                   fun base (car nodes1) (car nodes2))
       (cdr nodes1) (cdr nodes2))])]

   [(ast name fun base nodes)
    (cond
     [(null? nodes) base]
     [else
      (make-AST-for-all-node
       ast name fun (make-AST-node ast make-flow-graph-function-node name boolean=?
                                   fun base (car nodes))
       (cdr nodes))])]))

(define make-AST-for-all-non-bot-node
  (case-lambda
   [(ast name nodes)
    (cond
     [(null? nodes) (make-AST-const-node ast name #t)]
     [else (make-AST-for-all-non-bot-node
            ast name (make-AST-node ast make-flow-graph-function-node name boolean=?
                                    non-bot-type? (car nodes))
            (cdr nodes))])]
   [(ast name base nodes)
    (cond
     [(null? nodes) base]
     [else
      (make-AST-for-all-non-bot-node
         ast name (make-AST-node
                   ast make-flow-graph-function-node name boolean=?
                   (lambda (base node) (and base (non-bot-type? node)))
                   base (car nodes))
         (cdr nodes))])]))

(define (make-AST-const-node ast name const)
  (make-AST-node ast make-flow-graph-mutable-node name eq? const))
(define (make-AST-type-guard-node ast name guard node)
  (make-AST-node ast make-flow-graph-function-node name type=?
                 (lambda (guard val) (if guard val bot-type)) guard node))
(define (make-AST-const-type-guard-node ast name guard const)
  (make-AST-node ast make-flow-graph-function-node name type=?
                 (lambda (test) (if test const bot-type)) guard))

(define (make-AST-bitype-id-node ast name node)
  (make-AST-node ast make-flow-graph-function-node name bitype=?
                 (lambda (x) x) node))
(define (AST->type-id-node ast name expr)
  (AST-output-node expr))

(define (make-AST-enable-node ast name)
  (make-AST-node ast make-flow-graph-mutable-node name boolean=? #f))
(define (bool-node-enable! node) (flow-graph-mutable-node-set! node #t))

(define (make-AST-binding-node ast name)
  (make-AST-node ast make-flow-graph-mutable-node name type=? bot-type))

(define (make-AST-var-binding-node ast name var)
  (let ([binding-node
         (make-AST-binding-node
          ast (string-append name ":" (symbol->string (ident-symbol-name var))))])
    (ident-binding-node-set! var binding-node)
    (ident-preceding-ref-set! var (make-down-ref binding-node))
    binding-node))

(define (type-node-add! node value)
  (flow-graph-mutable-node-set!
   node (type-union value (flow-graph-node-output node))))

(define (make-AST-binding-node-add! ast name var node)
  (make-AST-node
   ast make-flow-graph-function-node name eq?
   (lambda (x) (type-node-add! (ident-binding-node var) x)) node))

(define (AST->bool-node ast name function expr)
  (make-AST-node ast make-flow-graph-function-node name boolean=? function
                 (AST-output-node expr)))

(define-record-type clause-info
  (nongenerative)
  (fields has-escaped in-called in-formals in-rest
          out-return out-return:rest-is-null out-return:rest-is-pair
          out-formals out-formals-bitop)
  (protocol
   (lambda (make)
     (lambda (clause
              has-rest has-escaped
              in-called in-formals&rest
              out-return out-formals&rest out-formals-bitop)

       (define (split-formals&rest list)
         (define (rec list)
           (cond
            [(null? (cdr list)) (values '() (car list))]
            [else
             (let-values ([(formals rest) (rec (cdr list))])
               (values (cons (car list) formals) rest))]))
         (cond
          [(not has-rest) (values list #f)]
          [else (rec list)]))

       (let-values
           ([(in-formals in-rest) (split-formals&rest in-formals&rest)]
            [(out-formals out-rest) (split-formals&rest out-formals&rest)])

         (define (make-out-return name pred?)
           (define (rest-return-type rest-bitype return-type)
             (type-filter
              return-type
              (pred? (bitype-true rest-bitype))
              (pred? (bitype-false rest-bitype))))
           (if (not out-rest)
               out-return
               (make-AST-node clause
                              make-flow-graph-function-node name type=?
                              rest-return-type out-rest out-return)))
         (make
             has-escaped in-called (list->vector in-formals) in-rest
             out-return
             (make-out-return "out-return&rest-is-null" null-type?)
             (make-out-return "out-return&rest-is-pair" pair-type?)
             (list->vector out-formals) out-formals-bitop))))))

(define-record-type up-ref (nongenerative) (fields ast node))
(define-record-type down-ref (nongenerative) (fields node))

(define-syntax define-proc-fun
  (syntax-rules ()
    [(_ (name clause-info formal-count arg-count args ...) body ...)
     (define (name table clauses missing-clause-node)
       (lambda (arg-count args ...)
         (let ([clause (case-lambda-clause-ref table clauses arg-count)])
           (if (not clause)
               missing-clause-node
               (let ([clause-info (case-lambda-clause-info clause)])
                 (let ([formal-count
                        (vector-length
                         (clause-info-out-formals clause-info))])
                   (assert (>= arg-count formal-count))
                   ;; clause-info formal-count arg-count args
                   body ...))))))]))

;; procedures for proc-type
(define (escape-proc has-escaped-node)
  (lambda () (bool-node-enable! has-escaped-node)))

(define-proc-fun (input-called-proc
                  clause-info formal-count arg-count callable)
  (when callable
    (let ([in-rest (clause-info-in-rest clause-info)])
      (when in-rest
        (type-node-add! in-rest
                        (cond
                         [(= arg-count formal-count) (type-of '())]
                         [(> arg-count formal-count) (type-of '(() . ()))]))))
    (bool-node-enable! (clause-info-in-called clause-info))))

(define-proc-fun (input-argument-proc
                  clause-info formal-count arg-count arg-position argument-type)
  (if (>= arg-position formal-count)
      (begin
        ;; values escape that go into rest arguments
        (assert (clause-info-in-rest clause-info))
        (mark-escaped-type! argument-type)
        '() ;; must return a value
        )
      (type-node-add!
       (vector-ref (clause-info-in-formals clause-info) arg-position)
       argument-type)))

(define-proc-fun (output-return-proc clause-info formal-count arg-count)
  (cond
   [(= arg-count formal-count) (clause-info-out-return:rest-is-null clause-info)]
   [(> arg-count formal-count) (clause-info-out-return:rest-is-pair clause-info)]))

(define-proc-fun (output-argument-proc
                  clause-info formal-count arg-count arg-position)
  (if (< arg-position formal-count)
      (vector-ref (clause-info-out-formals clause-info) arg-position)
      (clause-info-out-formals-bitop clause-info)))

;; AST * a -> ident or a
(define (AST-Ref->ident ast default)
  (AST-case ast
    [(Ref ident) ident]
    [else default]))

;; AST -> node (output-node or preceding-ref-node)
(define (AST->bare-var-node ast)
  (define var (AST-Ref->ident ast #f))
  ;; feed back if (1) it is a var and (2) it is an intersected
  ;; var, i.e., don't feed back singlely mentioned variables

  ;; NOTE: memq is O(1) b/c on a var the bubble length is at most one
  (if (and var (memq var (AST-bubble-vars ast)))
      (down-ref-node (ident-preceding-ref var))
      (AST-output-node ast)))

;; TODO: define turn-var-up/down in terms of single function
;; function must return a type
(define (turn-var-up! ast name function var)
  (define preceding-ref (ident-preceding-ref var))
  (define preceding-node (down-ref-node preceding-ref))
  (define node-name (string-append "turn-var-up:" name))
  (ident-preceding-ref-set!
   var (make-up-ref ast (make-AST-node ast make-flow-graph-function-node node-name
                                       bitype=? function preceding-node))))

;; function must return a bitype
(define (turn-var-down! ast name function var)
  (define preceding-ref (ident-preceding-ref var))
  (define preceding-node (up-ref-node preceding-ref))
  (define node-name (string-append "turn-var-down:" name ":"
                                   (symbol->string (ident-symbol-name var))))
  (ident-preceding-ref-set!
   var (make-down-ref (make-AST-node
                       ast make-flow-graph-function-node
                       node-name type=? function preceding-node))))

(define-syntax parallel-loop
  (syntax-rules ()
    [(_ ast-lists arguments action)
     (let loop ([asts ast-lists][args arguments][posts '()])
       (cond
        [(null? args) posts]
        [else
         (let* ([vars (AST-bubble-vars (car asts))]
                [pre (map* ident-preceding-ref vars)]
                [result (action (car args))] ;; ignore result
                [post (map* ident-preceding-ref vars)])
           (for-each* ident-preceding-ref-set! vars pre)
           (loop (cdr asts) (cdr args) (cons post posts)))]))]))

(define-syntax list-rev
  (syntax-rules ()
    [(_ acc) acc]
    [(_ acc x xs ...) (list-rev (cons x acc) xs ...)]))

(define-syntax parallel-list
  (syntax-rules ()
    [(_ (asts actions) ...)
     (list-rev
      '()
      (let* ([vars (AST-bubble-vars asts)]
             [pre (map* ident-preceding-ref vars)]
             [result actions] ;; ignore result
             [post (map* ident-preceding-ref vars)])
        (for-each* ident-preceding-ref-set! vars pre)
        post) ...)]))

(define-syntax fold-left* ;; equivalent to fold-left but ensures inlining
  (syntax-rules ()
    [(_ fun base args)
     (let loop ([b base]
                [a args])
       (cond
        [(null? a) b]
        [else (loop (fun b (car a)) (cdr a))]))]
    [(_ fun base args1 args2)
     (let loop ([b base]
                [a1 args1]
                [a2 args2])
       (cond
        [(null? a1) b]
        [else (loop (fun b (car a1) (car a2)) (cdr a1) (cdr a2))]))]))

(define-syntax for-each* ;; equivalent to for-each but ensures inlining
  (syntax-rules ()
    [(_ fun args)
     (let loop ([a args])
       (cond
        [(null? a) (values)]
        [else (fun (car a)) (loop (cdr a))]))]
    [(_ fun args1 args2)
     (let loop ([a1 args1]
                [a2 args2])
       (cond
        [(null? a1) (values)]
        [else (fun (car a1) (car a2)) (loop (cdr a1) (cdr a2))]))]
    [(_ fun args1 args2 args3)
     (let loop ([a1 args1]
                [a2 args2]
                [a3 args3])
       (cond
        [(null? a1) (values)]
        [else (fun (car a1) (car a2) (car a3))
              (loop (cdr a1) (cdr a2) (cdr a3))]))]))

(define-syntax map* ;; equivalent to map but ensures inlining
  (syntax-rules ()
    [(_ fun args)
     (let loop ([a args])
       (cond
        [(null? a) '()]
        [else (cons (fun (car a)) (loop (cdr a)))]))]
    [(_ fun args1 args2)
     (let loop ([a1 args1]
                [a2 args2])
       (cond
        [(null? a1) '()]
        [else (cons (fun (car a1) (car a2)) (loop (cdr a1) (cdr a2)))]))]))

(define-syntax parallel*
  (syntax-rules ()
    [(_ parent-ast* ident-up? node-up? eqv? first-ref combine asts action)
     (let ()
       (define parent-ast parent-ast*)

       ;; Get post values
       (define posts (reverse action))

       ;; Mark vars as not filled with a new value yet and collect the idents
       (define idents
         (fold-left*
          (lambda (idents ast)
            (fold-left*
             (lambda (idents ident)
               (if (ident-preceding-ref ident)
                   (begin
                     (ident-preceding-ref-set! ident #f)
                     (cons ident idents))
                   idents))
             idents
             (AST-bubble-vars ast)))
          '()
          asts))

       ;; Fill in vars & return list of uniq vars
       (for-each* (lambda (ast post)
                    (for-each*
                     (lambda (ident node)
                       (ident-preceding-ref-set!
                        ident
                        (cond
                          [(not combine) node]
                          [(not (ident-preceding-ref ident)) (first-ref node)]
                          [else
                           (let ([node (make-AST-node
                                        parent-ast make-flow-graph-function-node
                                        "parallel-uniq" eqv? combine
                                        ((if ident-up? up-ref-node down-ref-node)
                                         (ident-preceding-ref ident))
                                        ((if node-up? up-ref-node down-ref-node)
                                         node))])
                             (if ident-up?
                                 (make-up-ref parent-ast node)
                                 (make-down-ref node)))])))
                     (AST-bubble-vars ast) post))
                  asts posts)
       idents)]))

(define (parallel parent-ast up? eqv? first-ref combine asts thunks)
  (define (parallel-one ast thunk)
    (define vars (AST-bubble-vars ast))
    (define pre (map* ident-preceding-ref vars)) ;; save inner var nodes
    (define thunk-result (thunk)) ;; ignore result
    (define post (map* ident-preceding-ref vars)) ;; get inner var nodes
    (for-each* ident-preceding-ref-set! vars pre) ;; restore inner vars
    post)
  (define (add-node! ident node)
    (define field (if (eq? up? #t) up-ref-node down-ref-node))
    (ident-preceding-ref-set!
     ident
     (if (not (ident-preceding-ref ident))
         (first-ref node)
         (let ([node (make-AST-node
                      parent-ast make-flow-graph-function-node
                      "parallel-uniq" eqv? combine
                      (field (ident-preceding-ref ident))
                      ((if (eq? up? 'mixed)
                           up-ref-node field) node))])
           (if (eq? up? #t)
               (make-up-ref parent-ast node)
               (make-down-ref node))))))

  ;; Get post values
  (define posts (map* parallel-one asts thunks))

  ;; Mark vars as not filled with a new value yet and collect the idents
  (define idents
    (fold-left*
     (lambda (idents ast)
       (fold-left*
        (lambda (idents ident)
          (if (ident-preceding-ref ident)
              (begin
                (ident-preceding-ref-set! ident #f)
                (cons ident idents))
              idents))
        idents
        (AST-bubble-vars ast)))
     '()
     asts))

  ;; Fill in vars & return list of uniq vars
  (for-each* (lambda (ast post) (for-each* add-node! (AST-bubble-vars ast) post))
             asts posts)
  idents)

;; Runs rec-fun in parallel over asts
;; but bubbles get intersected as types (and then restored to bitypes)
;; and body*... gets to run over the types
(define-syntax parallel-intersected
  (syntax-rules ()
    [(_ parent-ast asts rec-fun reachable results idents)
     (begin
       (define idents
         (parallel*
          parent-ast #f #t type=?
          (lambda (node)
            (make-down-ref
             (make-AST-node
              parent-ast make-flow-graph-function-node
              "bitype->type" type=? bitype->type
              (up-ref-node node))))
          (lambda (node new) (type-intersect node (bitype->type new)))
          asts
          (parallel-loop asts asts rec-fun)))

       (define reachable
         ;; determine reachability for bottom vars (b/c might not
         ;; be bare) intersect variables that appear twice (stop if
         ;; any is bottom) (NOTE: singlely mentioned vars don't
         ;; need to be checked for bottom because they don't
         ;; intersect with each other.  If they were bottom then
         ;; the argument *value* would be bottom)
         (make-AST-for-all-non-bot-node
          parent-ast "reachable-by-env"

          ;; determine reachability from values
          (make-AST-for-all-non-bot-node
           parent-ast "reachable-by-value" (map* AST-output-node asts))

          (map* (lambda (ident) (down-ref-node (ident-preceding-ref ident)))
                idents)))

       (define (intersected-node ast)
         (define var (AST-Ref->ident ast #f))
         ;; feed back if (1) it is a var and (2) it is an intersected
         ;; var, i.e., don't feed back singlely mentioned variables
         ;; because they will be up-refs not down-refs

         ;; NOTE: memq is O(1) b/c on a var the bubble length is at most one
         (if (and var (memq var (AST-bubble-vars ast)))
             (down-ref-node (ident-preceding-ref var))
             (AST-output-node ast)))

       ;; feed back bare vars to formal positions
       (define results (map* intersected-node asts))
       (define-ignored
         (for-each* (lambda (ident)
                      (turn-var-up! parent-ast "parallel" type->bitype ident))
                    idents)))]))

(define (move-var-to-here ast var bypass-node)
  (define ref (ident-preceding-ref var))
  (define preceding-ast (up-ref-ast ref))
  (define preceding-node (up-ref-node ref))
  (if (eq? ast preceding-ast) ;; may already be here b/c of parallel
      preceding-node
      (let* ([env-flow-node
              (composition-stack-range
               (make-AST-const-node ast "tf-id" env-flow-function-identity)
               (AST-graph-context preceding-ast)
               (AST-graph-context ast))]
             [new-output-node
              (make-AST-node
               ast make-flow-graph-function-node
               (string-append "tf-apply:" (symbol->string (ident-symbol-name var)))
               bitype=? env-flow-function-apply
               env-flow-node preceding-node bypass-node)])
        (ident-preceding-ref-set! var (make-up-ref ast new-output-node))
        new-output-node)))

(define (AST-bypass-nodes ast)
  (define ast-bubble-vars (AST-bubble-vars ast))
  (define bypass-ref (map* ident-preceding-ref ast-bubble-vars))
  (assert (for-all down-ref? bypass-ref))
  (map* down-ref-node bypass-ref))

(define (make-value-flow-graph ast)

  (define bot-node ;; Used for return of arg-count mismatch
    (make-AST-const-node ast "top-node" bot-type))
  (define bibot-node ;; Used for out-formal of arg-count mismatch
    (make-AST-const-node ast "top-node" (type->bitype bot-type)))
  (define bitop-node ;; Used for arguments beyond 'rest'
    (make-AST-const-node ast "bitop-node" (type->bitype top-type)))

  (define (make-const-proc type)
    (make-proc-record
     (lambda () '())
     (lambda (arg-count callable) '())
     (lambda (arg-count position arg-type)
       (when (eq? type top-type) (mark-escaped-type! arg-type))
       '())
     (let ([node (make-AST-const-node ast "surrogate-proc-return" type)])
       (lambda (arg-count) node))
     (let ([node (make-AST-const-node
                  ast "surrogate-proc-arg" (type->bitype type))])
       (lambda (arg-count position) node))))
  (define top-proc (make-const-proc top-type))
  (define bot-proc (make-const-proc bot-type))
  (define (type->proc-record* arg-count)
    (lambda (reachable type)
      (select-CFA
       [(zero-CFA)
        (if (not reachable)
            ;; list-of bot-proc here so that we get the correct return
            ;; type out of the surrogate-fun calls
            (list bot-proc)
            (let ([proc* (type-procs type)])
              (if (null? proc*)
                  (list bot-proc)
                  (begin
                    (for-each
                     (lambda (proc)
                       (unless (eq? proc #t)
                         ((proc-record-input-called-proc proc) arg-count #t)))
                     proc*)
                    (map (lambda (proc) (if (eq? proc #t) top-proc proc))
                         proc*)))))]
       [(sub-zero-CFA)
        (if (not reachable)
            bot-proc
            (proc-type-bits-case (type-procs type)
              [(bot-proc-type-bits) bot-proc]
              [(top-proc-type-bits) top-proc]
              [(one-or-more-proc-type-bits proc)
               ((proc-record-input-called-proc proc) arg-count #t)
               proc]))])))

  (define (rec-true ast reachable)
    (define rec rec-true)
    (define ast-bubble-vars (AST-bubble-vars ast))
    (define bypass-nodes (AST-bypass-nodes ast))

    (AST-output-node-set!
     ast
     (AST-case ast
       [(Ref var)
        (let ([ref (ident-preceding-ref var)])
          (assert (down-ref? ref))
          ;; update preceding-ref to be here
          (unless (ident-assigned var)
            (turn-var-up! ast "ref" type->split-bitype var))
          (make-AST-type-guard-node ast "ref" reachable (down-ref-node ref)))]

       [(Quote obj)
        (make-AST-const-type-guard-node ast "quote" reachable (type-of obj))]

       [(Primref safe name)
        (make-AST-const-type-guard-node
         ast "Primref" reachable (prim-name->type ast name))]

       [(Set ident body)
        (rec body reachable)
        (make-AST-binding-node-add! ast "set!-var" ident (AST-output-node body))
        (make-AST-const-type-guard-node
         ast "set!" (AST->bool-node ast "set!-reach" non-bot-type? body) top-type)]

       [(Seq expr1 expr2)
        (rec expr1 reachable)
        (for-each* (lambda (x) (turn-var-down! ast "seq" bitype->type x))
                   (AST-bubble-vars expr1))
        (rec expr2 (AST->bool-node ast "reach" non-bot-type? expr1))
        (AST->type-id-node ast "seq" expr2)]

       [(If test conseq altern)
        (rec test reachable)

        (parallel*
         ast #t #t bitype=? (lambda (x) x) bitype-union (list conseq altern)
         (parallel-list
          [conseq
           (begin
             ;; Turn those mentioned in test and conseq back down conseq
             (for-each*
              (lambda (x) (turn-var-down! ast "conseq" bitype-true x))
              (AST-bubbles-intersect (AST-bubble-vars test)
                                     (AST-bubble-vars conseq)))
             (rec conseq (AST->bool-node ast "reach-conseq" true-type? test)))]
          [altern
           (begin
             ;; Turn those mentioned in test and altern back down altern
             (for-each*
              (lambda (x) (turn-var-down! ast "altern" bitype-false x))
              (AST-bubbles-intersect (AST-bubble-vars test)
                                     (AST-bubble-vars altern)))
             (rec altern (AST->bool-node ast "reach-altern" false-type? test)))]))

        (make-AST-node
         ast make-flow-graph-function-node "union" type=? type-union
         (AST-output-node conseq) (AST-output-node altern))]

       [(Letrec bindings body)
        (let ([lhs (map* letrec-binding-lhs (vector->list bindings))]
              [rhs (map* letrec-binding-rhs (vector->list bindings))])

          ;; Set-up bindings
          (define-ignored
            (for-each*
             (lambda (var) (make-AST-var-binding-node ast "letrec-binding" var))
             lhs))

          ;; intersect rhs
          (parallel-intersected
           ast rhs (lambda (rhs) (rec rhs reachable))
           body-reachable intersected-rhs idents)

          ;; Wire output of rhs to lhs bindings
          (for-each*
           (lambda (var node)
             (make-AST-binding-node-add! ast "letrec-binding-add" var node))
           lhs intersected-rhs)

          ;; Do the body
          (for-each*
           (lambda (var) (turn-var-down! ast "letrec" bitype->type var))
           (AST-bubble-vars body))
          (rec body body-reachable)

          ;; remove vars bound by letrec
          (for-each* ident-preceding-ref-delete! lhs)

          ;; return
          (AST->type-id-node ast "letrec-out" body))]

       [(CaseLambda table clauses)
        (let () ;; needed to allow 'define'
          (define (make-out-formal clause var in-formal)
            (define ref (ident-preceding-ref var))
            (if (down-ref? ref)
                ;; turn the var back up if it was never reffed
                (let ([node (down-ref-node ref)])
                  (assert (eq? in-formal node))
                  (make-AST-node
                   ast make-flow-graph-function-node
                   "unreffed-or-set!-var" bitype=? type->bitype node))
                ;; move all remaining vars to body result
                (make-AST-bitype-id-node
                 clause (string-append "out-formal:" (symbol->string
                                                      (ident-symbol-name var)))
                 (move-var-to-here (case-lambda-clause-body clause)
                                   var in-formal))))

          (define (calculate-clause clause)
            (define formal-idents (case-lambda-clause-formals clause))
            (define rest-ident (case-lambda-clause-rest clause))
            (define formals&rest-ident (case-lambda-clause-formals&rest clause))

            ;;;; input nodes
            (define in-called (make-AST-enable-node clause "in-called"))
            ;; setup vars bound by clause (don't set ast b/c it's down-var)
            (define in-formals&rest
              (map*
               (lambda (x) (make-AST-var-binding-node clause "lambda-binding" x))
               formals&rest-ident))

            ;;;; internal nodes
            (define in-reachable
              (make-AST-for-all-non-bot-node
               clause "in-formals-reachable" in-called in-formals&rest))

            ;; do body
            (define-ignored (rec (case-lambda-clause-body clause) in-reachable))

            ;;;; output nodes
            (define out-return
              (AST->type-id-node clause "case-lambda-return"
                                 (case-lambda-clause-body clause)))
            (define out-formals&rest
              (map* (lambda (var in-formal) (make-out-formal clause var in-formal))
                    formals&rest-ident in-formals&rest))

            ;; remove formal vars bound by clause
            (define-ignored
              (for-each* ident-preceding-ref-delete! formals&rest-ident))

            ;; Set the clause info
            (case-lambda-clause-info-set!
             clause
             (make-clause-info
              clause rest-ident has-escaped in-called in-formals&rest
              out-return out-formals&rest bitop-node)))

          (define has-escaped (make-AST-enable-node ast "has-escaped"))

          ;; We don't need to intersect the bubble vars (they are
          ;; always bypassed) but we must do a parallel to keep the up
          ;; vs down refs correct.

          (let* ([clauses (vector->list clauses)]
                 [asts (map* case-lambda-clause-body clauses)])
            (parallel* ast #t #t bitype=? #f #f asts
                       (parallel-loop
                        asts clauses
                        (lambda (clause)
                          (define body (case-lambda-clause-body clause))
                          (let ([bubble-vars (AST-bubble-vars body)]
                                [bypass-nodes (AST-bypass-nodes body)])
                            (calculate-clause clause)
                            (map* (lambda (x y) (move-var-to-here ast x y))
                                  bubble-vars bypass-nodes))))))

          (make-AST-node
           ast make-flow-graph-function-node "mark-escapes" eq?
           (lambda (has-escaped)
             (when has-escaped
               (for-each*
                (lambda (arg-count)
                  ((input-called-proc table clauses #f) arg-count #t)
                  (for-each*
                   (lambda (arg-position)
                     ((input-argument-proc table clauses #f)
                      arg-count arg-position top-type))
                   (iota arg-count)))
                (iota (vector-length table))))
             #f)
           has-escaped)
          (vector-for-each
           (lambda (clause)
             (make-AST-node
              ast make-flow-graph-function-node "out-return-mark-escapes" eq?
              (lambda (has-escaped out-return)
                (when has-escaped (mark-escaped-type! out-return))
                #f)
              has-escaped
              (clause-info-out-return (case-lambda-clause-info clause))))
           clauses)

          (make-AST-const-type-guard-node
           ast "case-lambda" reachable
           (make-proc-type
            (escape-proc has-escaped)
            (input-called-proc table clauses #f)
            (input-argument-proc table clauses #f)
            (output-return-proc table clauses bot-node)
            (output-argument-proc table clauses bibot-node))))]

       [(App fun actuals)
        (let () ;; needed to allow 'define'
          (define fun&actuals (cons fun (vector->list actuals)))
          (define arg-count (vector-length actuals))

          ;;;; Pre-call
          ;; intersect fun&actuals
          (parallel-intersected
           ast fun&actuals (lambda (expr) (rec expr reachable))
           call-reachable intersected-fun&actuals idents)

          ;; NOTE: also signals fun as callable
          (define surrogate-fun
            (make-AST-node
             ast make-flow-graph-function-node "surrogate-fun" proc-type-bits=?
             (type->proc-record* arg-count)
             call-reachable (car intersected-fun&actuals)))
          (define intersected-actuals (cdr intersected-fun&actuals))

          ;;;; Call entry

          ;; send arguments to function
          (define-ignored
            (for-each*
             (lambda (position intersected-actual)
               (make-AST-node
                ast make-flow-graph-function-node "input-argument" eq?
                (select-CFA
                 [(zero-CFA)
                  (lambda (proc-type* intersected-actual)
                    (for-each (lambda (proc-type)
                                ((proc-record-input-argument-proc proc-type)
                                 arg-count position intersected-actual))
                              proc-type*))]
                 [(sub-zero-CFA)
                  (lambda (proc-type intersected-actual)
                    ((proc-record-input-argument-proc proc-type)
                     arg-count position intersected-actual))])
                surrogate-fun intersected-actual))
             (iota arg-count)
             intersected-actuals))

          ;;;; Call return
          (define app-out-return
            (make-AST-node
             ast make-flow-graph-join-node "app-return" type=?
             (select-CFA
              [(zero-CFA)
               (lambda (proc-type*)
                 (when (null? proc-type*)
                   (error 'app-out-return "expected at least one proc-type"))
                 (fold-left
                  (lambda (node proc-type)
                    (make-AST-node
                     ast make-flow-graph-function-node "return-value" type=?
                     (lambda (type1 type2) (type-union type1 type2))
                     node ((proc-record-output-return-proc proc-type) arg-count)))
                  ((proc-record-output-return-proc (car proc-type*)) arg-count)
                  (cdr proc-type*)))]
              [(sub-zero-CFA)
               (lambda (proc-type)
                 ((proc-record-output-return-proc proc-type) arg-count))])
             surrogate-fun))

          (define app-out-formals
            (map*
             (lambda (position)
               (make-AST-node
                ast make-flow-graph-join-node "app-out-formal" bitype=?
                (select-CFA
                 [(zero-CFA)
                  (lambda (proc-type*)
                    (when (null? proc-type*)
                      (error 'app-out-return "expected at least one proc-type"))
                    (fold-left
                     (lambda (node proc-type)
                       (make-AST-node
                        ast make-flow-graph-function-node "arg-out-value" bitype=?
                        (lambda (bitype1 bitype2) (bitype-union bitype1 bitype2))
                        node ((proc-record-output-argument-proc proc-type)
                              arg-count position)))
                     ((proc-record-output-argument-proc (car proc-type*))
                      arg-count position)
                     (cdr proc-type*)))]
                 [(sub-zero-CFA)
                  (lambda (proc-type)
                    ((proc-record-output-argument-proc proc-type)
                     arg-count position))])
                surrogate-fun))
             (iota arg-count)))

          ;;;; Post-call

          ;; intersect formal output vars
          (define-ignored
            (for-each*
             (lambda (position actual formal)
               (let ([var (AST-Ref->ident actual #f)])
                 (when (and var (not (ident-assigned var)))
                   (ident-preceding-ref-set!
                    var
                    (make-up-ref
                     ast
                     (make-AST-node
                      ast make-flow-graph-function-node
                      "post-call-var-intersect" bitype=?
                      ;; NOTE: memq is O(1) b/c on a var the bubble length
                      ;; is at most one
                      (if (memq var (AST-bubble-vars actual))
                          ;; The var is mentioned multiple times thus the
                          ;; bitype has already been "void scrambled" by
                          ;; the parallel-intersect so we don't have to do
                          ;; that, but we new intersects need to be
                          ;; true/false specific.  We shouldn't do
                          ;; it because we might not be the first mention
                          ;; and we don't want to scramble info from
                          ;; other arguments.
                          bitype-intersect
                          ;; The var isn't mentioned multiple times so we
                          ;; have to do our own scrambling.  This is OK,
                          ;; since we are the only reference in this
                          ;; application.
                          (lambda (formal preceding)
                            (bitype-intersect
                             formal (bitype->void-bitype preceding))))
                      formal (up-ref-node (ident-preceding-ref var))))))))
             (iota (vector-length actuals))
             (vector->list actuals)
             app-out-formals))

          ;; determine post call return value
          (define (app-out-reachable-by-value base actual formal)
            (make-AST-node
             ast make-flow-graph-function-node "app-out-reachable-by-value" type=?
             (lambda (base actual formal)
               (type-filter
                base
                (non-bot-type? (type-intersect actual (bitype-true formal)))
                (non-bot-type? (type-intersect actual (bitype-false formal)))))
             base actual formal))

          (define (app-out-reachable-by-env base ident)
            (make-AST-node
             ast make-flow-graph-function-node "app-out-reachable-by-env" type=?
             (lambda (base bitype)
               (type-filter
                base
                (non-bot-type? (bitype-true bitype))
                (non-bot-type? (bitype-false bitype))))
             base
             (up-ref-node (ident-preceding-ref ident))))

          (fold-left*
           app-out-reachable-by-env
           (fold-left* app-out-reachable-by-value
                       app-out-return intersected-actuals app-out-formals)
           idents))]))

    ;; compute new nodes based on previous reference and bypass
    (for-each* (lambda (x y) (move-var-to-here ast x y))
               ast-bubble-vars bypass-nodes))

  (rec-true ast (make-AST-const-node ast "<top>" #t)))

)
